"""Tools for working with OMOP data."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._dataset import EHRDataset

__all__ = ["EHRDataset"]


def __getattr__(name: str):
    if name == "EHRDataset":
        from ._dataset import EHRDataset

        return EHRDataset
    msg = f"module '{__name__}' has no attribute '{name}'"
    raise AttributeError(msg)



import ipywidgets as widgets


def create_single_option_widget(title_text: str, options: list[str]):
    """Create a single-option selection widget (RadioButtons).

    Args:
        title_text: The title displayed above the radio buttons.
        options: List of strings representing the radio button options.
        default_value: The default selected value.

    Returns:
        tuple: A VBox containing the UI elements and the RadioButtons widget.
    """
    title = widgets.HTML(f"<h3>{title_text}</h3>")
    default_value = options[0]
    radio_buttons = widgets.RadioButtons(options=options, value=default_value, layout=widgets.Layout(width="auto"))
    ui = widgets.VBox([title, radio_buttons])

    return ui, radio_buttons


def create_multiple_options_widget(title_text: str, options: list[str]):
    """Create a multiple-option selection widget (SelectMultiple).

    Args:
        title_text: The title displayed above the selection box.
        options: List of strings representing the options.

    Returns:
        tuple: A VBox containing the UI elements and the SelectMultiple widget.
    """
    title = widgets.HTML(f"<h3 style='color: #333; font-family: Arial, sans-serif;'>{title_text}</h3>")

    height = f"{min(30 * len(options), 300)}px"

    select_multiple = widgets.SelectMultiple(
        options=options, value=(options[0],), layout=widgets.Layout(width="100%", height=height, font_family="Arial")
    )

    ui = widgets.VBox(
        [title, select_multiple],
        layout=widgets.Layout(padding="10px", border="1px solid #ddd", border_radius="5px", width="50%"),
    )

    return ui, select_multiple



from importlib.metadata import version

from . import dt, io, pl, tl
from ._feature_types import feature_type_overview, harmonize_missing_values, infer_feature_types, replace_feature_types
from .core import EHRData

__all__ = [
    "EHRData",
    "dt",
    "feature_type_overview",
    "harmonize_missing_values",
    "infer_feature_types",
    "io",
    "pl",
    "replace_feature_types",
    "tl",
]

__version__ = version("ehrdata")



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from ehrdata import EHRData
from lamin_utils import logger


def _cast_variables_to_float(edata: EHRData) -> None:
    """Cast the dtype of variables to float, and overwrite the values of the original arrays with the casted columns."""
    if edata.isbacked:
        msg = "Cannot cast variables to float when EHRData is backed."
        raise ValueError(msg)

    if edata.X is not None and not np.issubdtype(edata.X.dtype, np.number):
        for column in range(edata.shape[1]):
            with contextlib.suppress(ValueError):
                edata.X[:, column] = edata.X[:, column].astype(np.float64)

    for key in edata.layers:
        if edata.layers[key] is not None and not np.issubdtype(edata.layers[key].dtype, np.number):
            for column in range(edata.shape[1]):
                with contextlib.suppress(ValueError):
                    edata.layers[key][:, column] = edata.layers[key][:, column].astype(np.float64)


def _cast_arrays_dtype_to_float_or_str_if_nonnumeric_object(edata: EHRData) -> EHRData:
    """Cast the dtype of object arrays to float, and if this fails, to str."""
    if edata.X is not None and edata.X.dtype == np.object_:
        edata = edata.copy()
        try:
            edata.X = edata.X.astype(np.float64)
        except ValueError:
            edata.X = edata.X.astype(str)
        for layer, array in edata.layers.items():
            if array.dtype == np.object_:
                try:
                    edata.layers[layer] = array.astype(np.float64)
                except ValueError:
                    logger.warning(
                        f"edata.layers[{layer}] is of dtype {edata.layers[layer].dtype}: this is casted to dtype 'str' for saving to zarr."
                    )
                    edata.layers[layer] = array.astype(str)

    return edata



from __future__ import annotations

from functools import cache
from importlib.util import find_spec
from types import EllipsisType
from typing import TYPE_CHECKING
from warnings import warn

import h5py
import numpy as np
import pandas as pd
import scipy

#############################
# scipy sparse array compat #
#############################


CSMatrix = scipy.sparse.csr_matrix | scipy.sparse.csc_matrix
CSArray = scipy.sparse.csr_array | scipy.sparse.csc_array


class Empty:
    pass


Index1D = slice | int | str | np.int64 | np.ndarray | pd.Series
IndexRest = Index1D | EllipsisType
Index = (
    IndexRest
    | tuple[Index1D, IndexRest]
    | tuple[IndexRest, Index1D]
    | tuple[Index1D, Index1D, EllipsisType]
    | tuple[EllipsisType, Index1D, Index1D]
    | tuple[Index1D, EllipsisType, Index1D]
    | CSMatrix
    | CSArray
)
H5Group = h5py.Group
H5Array = h5py.Dataset
H5File = h5py.File


#############################
# Optional deps
#############################
@cache
def is_zarr_v2() -> bool:
    import zarr
    from packaging.version import Version

    return Version(zarr.__version__) < Version("3.0.0")


if is_zarr_v2():
    msg = "anndata will no longer support zarr v2 in the near future. Please prepare to upgrade to zarr>=3."
    warn(msg, DeprecationWarning, stacklevel=2)


if find_spec("awkward") or TYPE_CHECKING:
    import awkward  # noqa: F401
    from awkward import Array as AwkArray
else:

    class AwkArray:
        @staticmethod
        def __repr__():
            return "mock awkward.highlevel.Array"


if find_spec("zappy") or TYPE_CHECKING:
    from zappy.base import ZappyArray
else:

    class ZappyArray:
        @staticmethod
        def __repr__():
            return "mock zappy.base.ZappyArray"


if TYPE_CHECKING:
    # type checkers are confused and can only see …core.Array
    from dask.array.core import Array as DaskArray
elif find_spec("dask"):
    from dask.array import Array as DaskArray
else:

    class DaskArray:
        @staticmethod
        def __repr__():
            return "mock dask.array.core.Array"


# https://github.com/scverse/anndata/issues/1749
def is_cupy_importable() -> bool:
    try:
        import cupy  # noqa: F401
    except ImportError:
        return False
    return True


if is_cupy_importable() or TYPE_CHECKING:
    from cupy import ndarray as CupyArray
    from cupyx.scipy.sparse import csc_matrix as CupyCSCMatrix
    from cupyx.scipy.sparse import csr_matrix as CupyCSRMatrix
    from cupyx.scipy.sparse import spmatrix as CupySparseMatrix

    try:
        import dask.array as da
    except ImportError:
        pass
    else:
        da.register_chunk_type(CupyCSRMatrix)
        da.register_chunk_type(CupyCSCMatrix)
else:

    class CupySparseMatrix:
        @staticmethod
        def __repr__():
            return "mock cupyx.scipy.sparse.spmatrix"

    class CupyCSRMatrix:
        @staticmethod
        def __repr__():
            return "mock cupyx.scipy.sparse.csr_matrix"

    class CupyCSCMatrix:
        @staticmethod
        def __repr__():
            return "mock cupyx.scipy.sparse.csc_matrix"

    class CupyArray:
        @staticmethod
        def __repr__():
            return "mock cupy.ndarray"


def lazy_import_torch() -> None:
    try:
        import torch

        return torch
    except ImportError:
        msg = "The optional module 'torch' is not installed. Please install it using 'pip install ehrdata[torch]'."
        raise ImportError(msg) from None



import numpy as np
import pandas as pd
import pytest

from ehrdata import EHRData, feature_type_overview, harmonize_missing_values, infer_feature_types, replace_feature_types
from ehrdata.core.constants import MISSING_VALUES


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_harmonize_missing_values(sample_dataset, request):
    data, _ = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns))
    harmonize_missing_values(edata)

    for missing_value_string in MISSING_VALUES:
        assert missing_value_string not in edata.X.flatten()


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_harmonize_missing_values_layer(sample_dataset, request):
    data, _ = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    edata = EHRData(X=np.full(data.values.shape, "nan"), layers={"layer1": data.values})
    harmonize_missing_values(edata, layer="layer1")
    for missing_value_string in MISSING_VALUES:
        assert missing_value_string not in edata.layers["layer1"].flatten()


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_harmonize_missing_values_3D(sample_dataset, request):
    data, _ = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    R = data.values.reshape(2, -1, 2)
    edata = EHRData(R=R)
    harmonize_missing_values(edata, layer="R_layer")
    for missing_value_string in MISSING_VALUES:
        assert missing_value_string not in edata.R.flatten()


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_feature_type_inference_vanilla(sample_dataset, request):
    data, target_types = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns))
    infer_feature_types(edata)

    assert "feature_type" in edata.var.columns
    assert all(edata.var["feature_type"] == list(target_types.values()))


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_feature_type_inference_layer(sample_dataset, request):
    data, target_types = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    edata = EHRData(X=np.ones(data.values.shape), layers={"layer1": data.values})
    infer_feature_types(edata, layer="layer1")

    assert "feature_type" in edata.var.columns
    assert all(edata.var["feature_type"] == list(target_types.values()))


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_feature_type_inference_3D(sample_dataset, request):
    data, target_types = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    R = data.values.reshape(2, -1, 2)
    edata = EHRData(R=R)
    infer_feature_types(edata, layer="R_layer")

    assert "feature_type" in edata.var.columns
    assert all(edata.var["feature_type"] == list(target_types.values()))


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_feature_type_overview_vanilla(sample_dataset, request, capsys):
    data, _ = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns))
    feature_type_overview(edata)
    assert (
        " Detected feature types for EHRData object with 4 obs and 11 vars\n╠══ 📅 Date features\n╠══ 📐 Numerical features\n║   ╠══ 0\n║   ╠══ 1\n║   ╠══ 2\n║   ╠══ 3\n║   ╚══ 4\n╚══ 🗂️ Categorical features\n    ╠══ 10 (2 categories)\n    ╠══ 5 (4 categories)\n    ╠══ 6 (3 categories)\n    ╠══ 7 (2 categories)\n    ╠══ 8 (2 categories)\n    ╚══ 9 (2 categories)"
        in capsys.readouterr().out
    )


@pytest.mark.parametrize(
    "sample_dataset",
    [
        "variable_type_samples",
        "variable_type_samples_string_format",
    ],
)
def test_replace_feature_types(sample_dataset, request):
    data, target_types = request.getfixturevalue(sample_dataset)
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns).set_index(data.columns))
    infer_feature_types(edata)
    replace_feature_types(edata, ["int_column", "int_column_with_missing"], "categorical")

    target_types["int_column"] = "categorical"
    target_types["int_column_with_missing"] = "categorical"
    assert all(edata.var["feature_type"] == list(target_types.values()))


def test_replace_feature_types_not_inferred_raises_error(variable_type_samples):
    data, _ = variable_type_samples
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns).set_index(data.columns))

    with pytest.raises(ValueError):
        replace_feature_types(edata, ["int_column", "int_column_with_missing"], "categorical")


def test_replace_feature_types_invalid_type_raises_error(variable_type_samples):
    data, _ = variable_type_samples
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns).set_index(data.columns))
    infer_feature_types(edata)
    with pytest.raises(KeyError):
        replace_feature_types(edata, ["misspelt_column"], "categorical")


def test_replace_feature_types_unknown_feature_raises_error(variable_type_samples):
    data, _ = variable_type_samples
    data = pd.DataFrame(data)
    edata = EHRData(X=data.values, var=pd.DataFrame(data.columns).set_index(data.columns))
    infer_feature_types(edata)
    with pytest.raises(ValueError):
        replace_feature_types(edata, ["int_column"], "invalid_target")



import duckdb
import numpy as np
import pandas as pd
import pytest

import ehrdata as ed


def test_mimic_2():
    edata = ed.dt.mimic_2()
    assert edata.shape == (1776, 46, 0)
    expected_first_two_vars = ["aline_flg", "icu_los_day"]
    assert list(edata.var.index.values[:2]) == expected_first_two_vars
    expected_first_four_X = np.array([[1, 7.63], [0, 1.14]])
    assert np.array_equal(edata.X[:2, :2], expected_first_four_X)


def test_mimic_2_preprocessed():
    edata = ed.dt.mimic_2_preprocessed()
    assert edata.shape == (1776, 46, 0)
    expected_first_two_vars = ["ehrapycat_service_unit", "ehrapycat_day_icu_intime"]
    assert list(edata.var.index.values[:2]) == expected_first_two_vars
    expected_first_four_X = np.array([[2.0, 0.0], [1.0, 2.0]])
    assert np.array_equal(edata.X[:2, :2], expected_first_four_X)


def test_diabetes_130_raw():
    edata = ed.dt.diabetes_130_raw()
    assert edata.shape == (101766, 50, 0)
    expected_first_two_vars = ["encounter_id", "patient_nbr"]
    assert list(edata.var.index.values[:2]) == expected_first_two_vars
    expected_first_four_X = np.array([[2278392, 8222157], [149190, 55629189]])
    assert np.array_equal(edata.X[:2, :2], expected_first_four_X)


def test_diabetes_130_fairlearn():
    edata = ed.dt.diabetes_130_fairlearn()
    assert edata.shape == (101766, 24, 0)
    expected_first_two_vars = ["race", "gender"]
    assert list(edata.var.index.values[:2]) == expected_first_two_vars
    expected_first_four_X = np.array([["Caucasian", "Female"], ["Caucasian", "Female"]])
    assert np.array_equal(edata.X[:2, :2], expected_first_four_X)


@pytest.fixture
def duckdb_connection():
    """Fixture to create and return a DuckDB connection for testing."""
    con = duckdb.connect()
    yield con
    con.close()


def test_mimic_iv_omop():
    duckdb_connection = duckdb.connect()
    ed.dt.mimic_iv_omop(backend_handle=duckdb_connection)
    assert len(duckdb_connection.execute("SHOW TABLES").df()) == 30
    # sanity check of one table
    assert duckdb_connection.execute("SELECT * FROM person").df().shape == (100, 18)
    duckdb_connection.close()


def test_gibleed_omop():
    duckdb_connection = duckdb.connect()
    ed.dt.gibleed_omop(backend_handle=duckdb_connection)
    assert len(duckdb_connection.execute("SHOW TABLES").df()) == 36
    # sanity check of one table
    assert duckdb_connection.execute("SELECT * FROM person").df().shape == (2694, 18)
    duckdb_connection.close()


def test_synthea27nj_omop():
    duckdb_connection = duckdb.connect()
    ed.dt.synthea27nj_omop(backend_handle=duckdb_connection)
    assert len(duckdb_connection.execute("SHOW TABLES").df()) == 38
    # sanity check of one table
    assert duckdb_connection.execute("SELECT * FROM person").df().shape == (28, 18)
    duckdb_connection.close()


def test_physionet2012():
    edata = ed.dt.physionet2012()
    assert edata.shape == (11988, 37, 48)
    assert edata.tem.shape == (48, 1)
    assert edata.R.shape == (11988, 37, 48)
    assert edata.obs.shape == (11988, 10)
    assert edata.var.shape == (37, 1)

    # check a few hand-picked values for a stricter test
    # first entry set a
    assert edata.obs.loc["132539"].values[0] == "set-a"
    assert np.allclose(
        edata.obs.loc["132539"].values[1:].astype(np.float32),
        np.array([54.0, 0.0, -1.0, 4.0, 6, 1, 5, -1, 0], dtype=np.float32),
    )

    # first entry set b
    assert edata.obs.loc["142675"].values[0] == "set-b"
    assert np.allclose(
        edata.obs.loc["142675"].values[1:].astype(np.float32),
        np.array([70.0, 1.0, 175.3, 2.0, 27, 14, 9, 7, 1], dtype=np.float32),
    )

    # first entry set c
    assert edata.obs.loc["152871"].values[0] == "set-c"
    assert np.allclose(
        edata.obs.loc["152871"].values[1:].astype(np.float32),
        np.array([71.0, 1.0, 167.6, 4.0, 19, 10, 23, -1, 0], dtype=np.float32),
    )

    # first entry c two different HR value
    assert np.isclose(edata[edata.obs.index.get_loc("152871"), "HR", 0].R.item(), 65)
    assert np.isclose(edata[edata.obs.index.get_loc("152871"), "HR", 28].R.item(), 68)


def test_physionet2012_arguments():
    edata = ed.dt.physionet2012(
        interval_length_number=2,
        interval_length_unit="min",
        num_intervals=24,
        aggregation_strategy="first",
        drop_samples=None,
    )
    assert edata.shape == (12000, 37, 24)
    assert edata.tem.shape == (24, 1)
    assert edata.R.shape == (12000, 37, 24)
    assert edata.obs.shape == (12000, 10)
    assert edata.var.shape == (37, 1)


@pytest.mark.parametrize("sparse_param", [False])  # [False, True]
def test_ehrdata_blobs(sparse_param):
    """Test the ehrdata_blobs function."""
    edata = ed.dt.ehrdata_blobs(n_observations=100, n_variables=5, base_timepoints=10, sparse=sparse_param)

    assert isinstance(edata, ed.EHRData)

    assert edata.shape == (100, 5, 10)
    assert edata.n_obs == 100
    assert edata.n_vars == 5
    assert edata.n_t == 10

    # Test X data
    if not sparse_param:
        assert isinstance(edata.X, np.ndarray)
        assert edata.X.shape == (100, 5)
    # else:
    #     assert sparse.issparse(ehr_data.X)
    #     assert ehr_data.X.shape == (100, 5)

    # Test R data
    if not sparse_param:
        assert isinstance(edata.R, np.ndarray)
        assert edata.R.shape == (100, 5, 10)
    # else:
    #     from sparse import COO
    #     assert isinstance(ehr_data.R, COO)
    #     assert ehr_data.R.shape == (100, 5, 10)

    # Test obs DataFrame
    assert isinstance(edata.obs, pd.DataFrame)
    assert "cluster" in edata.obs.columns
    assert edata.obs.shape == (100, 1)

    # Test var DataFrame
    assert isinstance(edata.var, pd.DataFrame)
    assert edata.var.shape == (5, 0)

    # Test t DataFrame
    assert isinstance(edata.tem, pd.DataFrame)
    assert "timepoint" in edata.tem.columns
    assert edata.tem.shape == (10, 2)


def test_ehrdata_blobs_distribution():
    edata = ed.dt.ehrdata_blobs(
        n_observations=500,
        n_variables=10,
        n_centers=3,
        base_timepoints=15,
        cluster_std=0.5,
        sparse=False,
        random_state=42,
    )

    assert isinstance(edata, ed.EHRData)
    assert edata.shape == (500, 10, 15)

    clusters = edata.obs["cluster"].astype(int).values

    # Test cluster separation in X
    for cluster_id in np.unique(clusters):
        cluster_mask = clusters == cluster_id
        cluster_points = edata.X[cluster_mask]

        cluster_center = np.mean(cluster_points, axis=0)

        # Calculate average distance to center
        distances = np.linalg.norm(cluster_points - cluster_center, axis=1)
        avg_distance = np.mean(distances)

        # Check that points are reasonably clustered (average distance should be near cluster_std)
        expected_distance = 0.5 * np.sqrt(10)  # cluster_std * sqrt(dimensions)
        assert 0.3 * expected_distance < avg_distance < 3.0 * expected_distance

    # Test time evolution in R
    # Check that variation increases with time
    time_variations = []
    for t in range(edata.n_t):
        time_slice = edata.R[:, :, t]
        variation = np.std(time_slice)
        time_variations.append(variation)

    # Verify increasing variation trend
    assert time_variations[-1] > time_variations[0]

    # Test that R at t=0 is close to X
    first_timepoint = edata.R[:, :, 0]
    correlation = np.corrcoef(first_timepoint.flatten(), edata.X.flatten())[0, 1]
    assert correlation > 0.5


def test_ehrdata_ts_blobs_irregular():
    edata = ed.dt.ehrdata_blobs(
        n_observations=300,
        n_variables=8,
        n_centers=4,
        base_timepoints=20,
        cluster_std=0.5,
        sparse=False,
        variable_length=True,
        time_shifts=True,
        seasonality=True,
        irregular_sampling=True,
        missing_values=0.1,
        random_state=42,
    )

    assert isinstance(edata, ed.EHRData)
    assert edata.n_obs == 300
    assert edata.n_vars == 8

    # Check that time dimension exists and has more points than base_timepoints
    assert edata.n_t > 20
    assert "time_value" in edata.tem.columns

    # Test for irregular time sampling
    time_values = edata.tem["time_value"].values
    time_diffs = np.diff(time_values)
    # Time differences should have meaningful variation with irregular sampling
    assert np.std(time_diffs) > 0.001

    # Test for missing values in R
    nan_count = np.isnan(edata.R).sum()
    total_elements = np.prod(edata.R.shape)
    missing_ratio = nan_count / total_elements
    assert missing_ratio > 0.05

    # Test for variable length time series
    valid_counts = []
    for i in range(edata.n_obs):
        valid_count = np.sum(~np.isnan(edata.R[i, 0, :]))
        valid_counts.append(valid_count)

    valid_counts = np.array(valid_counts)
    # Check for variation in time series lengths
    assert np.std(valid_counts) > 0.5
    assert np.max(valid_counts) > np.min(valid_counts)

    # Test for time shifts
    clusters = edata.obs["cluster"].astype(int).values

    shift_detected = False
    for cluster in np.unique(clusters):
        cluster_mask = clusters == cluster
        obs_indices = np.where(cluster_mask)[0]

        if len(obs_indices) >= 2:
            obs1 = obs_indices[0]
            obs2 = obs_indices[1]

            valid_times1 = np.where(~np.isnan(edata.R[obs1, 0, :]))[0]
            valid_times2 = np.where(~np.isnan(edata.R[obs2, 0, :]))[0]

            if len(valid_times1) > 0 and len(valid_times2) > 0:
                first_time1 = valid_times1[0]
                first_time2 = valid_times2[0]

                if abs(first_time1 - first_time2) > 0:
                    shift_detected = True
                    break

    assert shift_detected

    # Test for seasonality
    seasonal_pattern_found = False
    for i in range(min(10, edata.n_obs)):  # Check first 10 observations max
        for v in range(edata.n_vars):
            values = edata.R[i, v, :]
            valid_mask = ~np.isnan(values)

            if np.sum(valid_mask) >= 10:  # Need at least 10 valid points
                time_series = values[valid_mask]

                # Look for patterns by checking if the time series is non-monotonic
                # True seasonality will have ups and downs
                diffs = np.diff(time_series)
                sign_changes = np.sum((diffs[:-1] * diffs[1:]) < 0)

                # If we have multiple sign changes, there's likely a pattern
                if sign_changes >= 3:
                    seasonal_pattern_found = True
                    break
        if seasonal_pattern_found:
            break

    assert seasonal_pattern_found



from __future__ import annotations

from typing import Literal


def get_table_catalog_dict(version: Literal["5.4"] = "5.4") -> dict[str, str]:
    """Get the table catalog dictionary of the OMOP CDM.

    Args:
        version:
            The version of the OMOP CDM. Currently, only 5.4 is supported.

    Returns:
        A dictionary of the table catalog. The key is the category of the table, and the value is a list of table names.
    """
    table_catalog_dict = {}
    table_catalog_dict["Clinical data"] = [
        "person",
        "observation_period",
        "specimen",
        "death",
        "visit_occurrence",
        "visit_detail",
        "procedure_occurrence",
        "drug_exposure",
        "device_exposure",
        "condition_occurrence",
        "measurement",
        "note",
        "note_nlp",
        "observation",
        "fact_relationship",
        "episode",
    ]

    table_catalog_dict["Health system data"] = ["location", "care_site", "provider"]
    table_catalog_dict["Health economics data"] = ["payer_plan_period", "cost"]
    table_catalog_dict["Standardized derived elements"] = [
        "cohort",
        "cohort_definition",
        "drug_era",
        "dose_era",
        "condition_era",
    ]
    table_catalog_dict["Metadata"] = ["cdm_source", "metadata"]
    table_catalog_dict["Vocabulary"] = [
        "concept",
        "vocabulary",
        "domain",
        "concept_class",
        "concept_relationship",
        "relationship",
        "concept_synonym",
        "concept_ancestor",
        "source_to_concept_map",
        "drug_strength",
    ]

    return table_catalog_dict



from pathlib import Path

import anndata as ad
import dask.array as da
import duckdb
import numpy as np
import pandas as pd
import pytest
import sparse as sp

from ehrdata import EHRData
from ehrdata.core.constants import R_LAYER_KEY
from ehrdata.io.omop import setup_connection


def _assert_shape_matches(
    edata: EHRData, shape: tuple[int, int, int], *, check_X_None: bool = False, check_R_None: bool = False
):
    assert edata.shape == shape

    if check_X_None:
        assert edata.X is None
    else:
        assert edata.X.shape == shape[0:2]

    if check_R_None:
        assert edata.R is None
    else:
        assert edata.R.shape == shape

    assert isinstance(edata.obs, pd.DataFrame)
    assert len(edata.obs) == shape[0]
    assert edata.n_obs == shape[0]

    assert isinstance(edata.var, pd.DataFrame)
    assert len(edata.var) == shape[1]
    assert edata.n_vars == shape[1]

    assert isinstance(edata.tem, pd.DataFrame)
    assert len(edata.tem) == shape[2]
    assert edata.n_t == shape[2]

    for key in edata.layers:
        if key != R_LAYER_KEY:
            assert edata.layers[key].shape == shape[0:2]


def _assert_dtype_object_array_with_missing_values_equal(a: np.ndarray, b: np.ndarray):
    # need to use pd.isnull to check for np.isnan in dtype object arrays, because np.isnan does not work on dtype object array
    a = a.copy()
    b = b.copy()
    assert np.array_equal(pd.isnull(a), pd.isnull(b))
    # if verified equal position of nan values, replace with 0 and verify the rest of the entries are equal
    a[pd.isnull(a)] = 0
    b[pd.isnull(b)] = 0
    assert np.array_equal(a, b)


@pytest.fixture
def csv_basic():
    return pd.read_csv("tests/data/toy_csv/csv_basic.csv")


@pytest.fixture
def csv_non_num_with_missing():
    return pd.read_csv("tests/data/toy_csv/csv_non_num_with_missing.csv")


@pytest.fixture
def csv_num_with_missing():
    return pd.read_csv("tests/data/toy_csv/csv_num_with_missing.csv")


@pytest.fixture
def X_numpy_32():
    return np.arange(1, 7).reshape(3, 2)


@pytest.fixture
def X_numpy_33():
    return np.arange(1, 10).reshape(3, 3)


@pytest.fixture
def R_numpy_322():
    return np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]], [[9, 10], [11, 12]]])


@pytest.fixture
def R_sparse_322():
    return sp.COO(np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]], [[9, 10], [11, 12]]]))


@pytest.fixture
def R_dask_322():
    return da.from_array(np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]], [[9, 10], [11, 12]]]), chunks=(3, 2, 2))


@pytest.fixture
def R_numpy_333():
    return np.arange(1, 28).reshape(3, 3, 3)


@pytest.fixture
def obs_31():
    return pd.DataFrame({"obs_col_1": [1, 2, 3]}, index=["obs1", "obs2", "obs3"])


@pytest.fixture
def var_21():
    return pd.DataFrame({"var_col_1": [1, 2]}, index=["var1", "var2"])


@pytest.fixture
def var_31():
    return pd.DataFrame({"var_col_1": [1, 2, 3]}, index=["var1", "var2", "var3"])


@pytest.fixture
def tem_11():
    return pd.DataFrame({"tem_col_1": [1]}, index=["t1"])


@pytest.fixture
def tem_21():
    return pd.DataFrame({"tem_col_1": [1, 2]}, index=["t1", "t2"])


@pytest.fixture
def tem_31():
    return pd.DataFrame({"tem_col_1": [1, 2, 3]}, index=["t1", "t2", "t3"])


@pytest.fixture
def edata_333(X_numpy_33, R_numpy_333, obs_31, var_31, tem_31):
    return EHRData(X=X_numpy_33, R=R_numpy_333, obs=obs_31, var=var_31, tem=tem_31)


@pytest.fixture
def edata_330(X_numpy_33, obs_31, var_31):
    return EHRData(X=X_numpy_33, obs=obs_31, var=var_31)


@pytest.fixture
def adata_33(X_numpy_33, obs_31, var_31):
    return ad.AnnData(X=X_numpy_33, obs=obs_31, var=var_31)


@pytest.fixture
def variable_type_samples():
    column_types = {
        "float_column": np.array([1.1, 1.2, 1.3, 2.1]),
        "float_column_with_missing": np.array([1.1, np.nan, 1.3, 2.1]),
        "int_column": np.array([1, 2, 3, 4]),
        "int_column_with_missing": np.array([1, np.nan, 3, 4]),
        "int_column_irregular": np.array([1, 2, 5, 6]),
        "string_column": np.array(["a", "b", "c", "d"]),
        "string_column_with_missing": np.array(["a", np.nan, "c", "d"]),
        "string_column_with_missing_strings": np.array(["a", "np.nan", "nan", "d"]),
        "bool_column_TrueFalse": np.array([True, False, True, False]),
        "bool_column_01": np.array([1, 0, 1, 0]),
        "bool_column_with_missing": np.array([True, np.nan, True, False]),
    }

    target_types = {
        "float_column": "numeric",
        "float_column_with_missing": "numeric",
        "int_column": "numeric",
        "int_column_with_missing": "numeric",
        "int_column_irregular": "numeric",
        "string_column": "categorical",
        "string_column_with_missing": "categorical",
        "string_column_with_missing_strings": "categorical",
        "bool_column_TrueFalse": "categorical",
        "bool_column_01": "categorical",
        "bool_column_with_missing": "categorical",
    }
    return column_types, target_types


@pytest.fixture
def variable_type_samples_string_format(variable_type_samples):
    # cast entries with .astype(str)
    data, target_types = variable_type_samples
    for key, value in data.items():
        data[key] = value.astype(str)
    return data, target_types


@pytest.fixture
def edata_nonnumeric_missing_330(obs_31, var_31):
    # create X of dtype object - np would create a string array
    X = pd.DataFrame(
        [
            [3, "E10", 12.1],
            [np.nan, "E11", 13.2],
            [14, np.nan, 12.5],
        ]
    ).to_numpy()
    return EHRData(X=X, layers={"other_layer": X}, obs=obs_31, var=var_31)


@pytest.fixture
def edata_basic_with_tem_full():
    edata_basic_with_tem_dict = {
        "X": np.ones((5, 4)),
        "R": np.ones((5, 4, 2)),
        "obs": pd.DataFrame({"survival": [1, 2, 3, 4, 5]}),
        "var": pd.DataFrame({"variables": ["var_1", "var_2", "var_3", "var_4"]}),
        "obsm": {"obs_level_representation": np.ones((5, 2))},
        "varm": {"var_level_representation": np.ones((4, 2))},
        "layers": {"other_layer": np.ones((5, 4))},
        "obsp": {"obs_level_connectivities": np.ones((5, 5))},
        "varp": {"var_level_connectivities": np.random.randn(4, 4)},
        "uns": {"information": ["info1"]},
        "tem": pd.DataFrame({"timestep": ["t1", "t2"]}),
    }
    return EHRData(**edata_basic_with_tem_dict)


@pytest.fixture
def omop_connection_vanilla():
    con = duckdb.connect()
    setup_connection(path="tests/data/toy_omop/vanilla", backend_handle=con)
    yield con
    con.close()


@pytest.fixture
def omop_connection_capital_letters():
    con = duckdb.connect()
    setup_connection(path="tests/data/toy_omop/capital_letters", backend_handle=con)
    yield con
    con.close()


@pytest.fixture
def omop_connection_empty_observation():
    con = duckdb.connect()
    setup_connection(path="tests/data/toy_omop/empty_observation", backend_handle=con)
    yield con
    con.close()


@pytest.fixture
def omop_connection_multiple_units():
    con = duckdb.connect()
    setup_connection(path="tests/data/toy_omop/multiple_units", backend_handle=con)
    yield con
    con.close()


def _assert_io_read(edata: EHRData):
    """Assert the test zarr and h5ad files are read correctly."""
    assert "survival" in edata.obs.columns
    assert all(edata.obs["survival"].values == [1, 2, 3, 4, 5])
    assert "variables" in edata.var.columns
    assert all(edata.var["variables"].values == ["var_1", "var_2", "var_3", "var_4"])
    assert "obs_level_representation" in edata.obsm
    assert edata.obsm["obs_level_representation"].shape == (5, 2)
    assert "var_level_representation" in edata.varm
    assert edata.varm["var_level_representation"].shape == (4, 2)
    # shapes are enforced by AnnData/EHRData for the below, no need to test
    assert "other_layer" in edata.layers
    assert "obs_level_connectivities" in edata.obsp
    assert "var_level_connectivities" in edata.varp
    assert "information" in edata.uns


TEST_DATA_PATH = Path(__file__).parent / "data"



#########################
# Toy zarr files creation
#########################
import anndata as ad
import h5py
import numpy as np
import pandas as pd
import scipy as sp
import zarr

#########################
# adata_basic.zarr/h5ad:  basic AnnData object

adata_basic = ad.AnnData(
    X=np.ones((5, 4)),
    obs=pd.DataFrame({"survival": [1, 2, 3, 4, 5]}),
    var=pd.DataFrame({"variables": ["var_1", "var_2", "var_3", "var_4"]}),
    obsm={"obs_level_representation": np.ones((5, 2))},
    varm={"var_level_representation": np.ones((4, 2))},
    layers={"other_layer": np.ones((5, 4))},
    obsp={"obs_level_connectivities": np.ones((5, 5))},
    varp={"var_level_connectivities": np.random.randn(4, 4)},
    uns={"information": ["info1"]},
)
# adata_basic.write_zarr("toy_zarr/adata_basic.zarr")
# adata_basic.write_h5ad("toy_h5ad/adata_basic.h5ad")

#########################
# edata_basic_with_tem.zarr/h5ad:  basic EHRData object with tem, R


edata_basic_with_tem_dict = {
    "X": np.ones((5, 4)),
    "R": np.ones((5, 4, 2)),
    "obs": pd.DataFrame({"survival": [1, 2, 3, 4, 5]}),
    "var": pd.DataFrame({"variables": ["var_1", "var_2", "var_3", "var_4"]}),
    "obsm": {"obs_level_representation": np.ones((5, 2))},
    "varm": {"var_level_representation": np.ones((4, 2))},
    "layers": {"other_layer": np.ones((5, 4))},
    "obsp": {"obs_level_connectivities": np.ones((5, 5))},
    "varp": {"var_level_connectivities": np.random.randn(4, 4)},
    "uns": {"information": ["info1"]},
    "tem": pd.DataFrame({"timestep": ["t1", "t2"]}),
}
with zarr.open("toy_zarr/edata_basic_with_tem.zarr", "w") as zarr_file:
    for k, v in edata_basic_with_tem_dict.items():
        ad.io.write_elem(zarr_file, k, v)


with h5py.File("toy_h5ad/edata_basic_with_tem.h5ad", "w") as h5ad_file:
    for k, v in edata_basic_with_tem_dict.items():
        ad.io.write_elem(h5ad_file, k, v)


#########################
# edata_sparse_with_tem.zarr/h5ad:  basic EHRData object with tem, R, sparse X
edata_sparse_with_tem_dict = edata_basic_with_tem_dict.copy()
edata_sparse_with_tem_dict["X"] = sp.sparse.csr_matrix(edata_sparse_with_tem_dict["X"])
edata_sparse_with_tem_dict["layers"]["other_layer"] = sp.sparse.csr_matrix(
    edata_sparse_with_tem_dict["layers"]["other_layer"]
)
with zarr.open("toy_zarr/edata_sparse_with_tem.zarr", "w") as zarr_file:
    for k, v in edata_sparse_with_tem_dict.items():
        ad.io.write_elem(zarr_file, k, v)


with h5py.File("toy_h5ad/edata_sparse_with_tem.h5ad", "w") as h5ad_file:
    for k, v in edata_sparse_with_tem_dict.items():
        ad.io.write_elem(h5ad_file, k, v)



from collections.abc import Iterable
from functools import singledispatch

import numpy as np
import pandas as pd
from anndata.compat import Index


@singledispatch
def _subset(a, subset_idx):
    msg = f"Subsetting of type {type(a)} in field .r is not implemented."
    raise NotImplementedError(msg)


@_subset.register(np.ndarray)
@_subset.register(pd.DataFrame)
def _(a: np.ndarray | pd.DataFrame, subset_idx: Index):
    # Select as combination of indexes, not coordinates
    # Correcting for indexing behaviour of np.ndarray
    if all(isinstance(x, Iterable) for x in subset_idx):
        subset_idx = np.ix_(*subset_idx)
    return a[subset_idx]



import anndata as ad
import numpy as np
import pandas as pd
import pytest
import zarr
from scipy.sparse import issparse
from tests.conftest import (
    TEST_DATA_PATH,
    _assert_dtype_object_array_with_missing_values_equal,
    _assert_io_read,
    _assert_shape_matches,
)

from ehrdata.io import read_zarr, write_zarr

TEST_PATH_ZARR = TEST_DATA_PATH / "toy_zarr"


@pytest.mark.parametrize("harmonize_missing_values", [False, True])
@pytest.mark.parametrize("cast_variables_to_float", [False, True])
def test_read_zarr_basic(harmonize_missing_values, cast_variables_to_float):
    edata = read_zarr(
        filename=TEST_PATH_ZARR / "adata_basic.zarr",
        harmonize_missing_values=harmonize_missing_values,
        cast_variables_to_float=cast_variables_to_float,
    )

    _assert_shape_matches(edata, (5, 4, 0), check_R_None=True)
    _assert_io_read(edata)


@pytest.mark.parametrize("harmonize_missing_values", [False, True])
@pytest.mark.parametrize("cast_variables_to_float", [False, True])
def test_read_zarr_basic_with_tem(harmonize_missing_values, cast_variables_to_float):
    edata = read_zarr(
        filename=TEST_PATH_ZARR / "edata_basic_with_tem.zarr",
        harmonize_missing_values=harmonize_missing_values,
        cast_variables_to_float=cast_variables_to_float,
    )
    _assert_shape_matches(edata, (5, 4, 2))
    _assert_io_read(edata)

    assert "timestep" in edata.tem.columns
    assert all(edata.tem["timestep"].values == ["t1", "t2"])


@pytest.mark.parametrize("harmonize_missing_values", [False, True])
@pytest.mark.parametrize("cast_variables_to_float", [False, True])
def test_read_zarr_sparse_with_tem(harmonize_missing_values, cast_variables_to_float):
    edata = read_zarr(
        filename=TEST_PATH_ZARR / "edata_sparse_with_tem.zarr",
        harmonize_missing_values=harmonize_missing_values,
        cast_variables_to_float=cast_variables_to_float,
    )
    _assert_shape_matches(edata, (5, 4, 2))
    _assert_io_read(edata)

    assert "timestep" in edata.tem.columns
    assert all(edata.tem["timestep"].values == ["t1", "t2"])

    assert issparse(edata.X)
    assert issparse(edata.layers["other_layer"])


@pytest.mark.parametrize("edata_name", ["edata_333", "edata_basic_with_tem_full", "edata_nonnumeric_missing_330"])
def test_write_zarr_basic(edata_name, request, tmp_path):
    edata = request.getfixturevalue(edata_name)

    write_zarr(edata, f"{tmp_path}/{edata_name}.zarr")

    zarr_file = zarr.open(f"{tmp_path}/{edata_name}.zarr", mode="r")
    # Note that R is not included in the list because it is just a value of the .layers field
    assert set(dict(zarr_file).keys()) == {
        "X",
        "obs",
        "var",
        "obsm",
        "varm",
        "layers",
        "obsp",
        "varp",
        "raw",
        "uns",
        "tem",
    }

    assert np.array_equal(ad.io.read_elem(zarr_file["X"]).astype(str), edata.X.astype(str))

    pd.testing.assert_frame_equal(ad.io.read_elem(zarr_file["obs"]), edata.obs)
    pd.testing.assert_frame_equal(ad.io.read_elem(zarr_file["var"]), edata.var)
    pd.testing.assert_frame_equal(ad.io.read_elem(zarr_file["tem"]), edata.tem)
    for key in edata.obsm:
        assert key in ad.io.read_elem(zarr_file["obsm"])
        assert np.array_equal(ad.io.read_elem(zarr_file["obsm"][key]), edata.obsm[key])
    for key in edata.varm:
        assert key in ad.io.read_elem(zarr_file["varm"])
        assert np.array_equal(ad.io.read_elem(zarr_file["varm"][key]), edata.varm[key])
    for key in edata.obsp:
        assert key in ad.io.read_elem(zarr_file["obsp"])
        assert np.array_equal(ad.io.read_elem(zarr_file["obsp"][key]), edata.obsp[key])
    for key in edata.varp:
        assert key in ad.io.read_elem(zarr_file["varp"])
        assert np.array_equal(ad.io.read_elem(zarr_file["varp"][key]), edata.varp[key])
    for key in edata.uns:
        assert key in ad.io.read_elem(zarr_file["uns"])


@pytest.mark.parametrize("edata_name", ["edata_333", "edata_basic_with_tem_full", "edata_nonnumeric_missing_330"])
def test_write_read_zarr_basic(edata_name, request, tmp_path):
    edata = request.getfixturevalue(edata_name)
    write_zarr(edata.copy(), f"{tmp_path}/{edata_name}.zarr")
    edata_read = read_zarr(f"{tmp_path}/{edata_name}.zarr")

    assert edata.shape == edata_read.shape

    _assert_dtype_object_array_with_missing_values_equal(edata.X, edata_read.X)
    for key in edata.layers:
        _assert_dtype_object_array_with_missing_values_equal(edata.layers[key], edata_read.layers[key])

    pd.testing.assert_frame_equal(edata.obs, edata_read.obs)
    pd.testing.assert_frame_equal(edata.var, edata_read.var)
    pd.testing.assert_frame_equal(edata.tem, edata_read.tem)
    for key in edata.obsm:
        assert key in edata_read.obsm
        assert np.array_equal(edata.obsm[key], edata_read.obsm[key])
    for key in edata.varm:
        assert key in edata_read.varm
        assert np.array_equal(edata.varm[key], edata_read.varm[key])
    for key in edata.obsp:
        assert key in edata_read.obsp
        assert np.array_equal(edata.obsp[key], edata_read.obsp[key])
    for key in edata.varp:
        assert key in edata_read.varp
        assert np.array_equal(edata.varp[key], edata_read.varp[key])
    for key in edata.uns:
        assert key in edata_read.uns
        assert np.array_equal(edata.uns[key], edata_read.uns[key])



import numpy as np
import pytest
from anndata import AnnData

from ehrdata.pl.vitessce import gen_config


@pytest.fixture
def adata() -> AnnData:
    return AnnData(
        X=np.array([[1, 2, 3], [4, 5, 6]]),
        obs={"gender_concept_id": ["M", "F"]},
        obsm={"X_pca": np.array([[1, 2], [3, 4]])},
    )


def test_gen_config(adata, tmp_path):
    adata.write_zarr(path := tmp_path / "test.zarr")
    gen_config(path)


# needs more setup until it works
# def test_gen_config_lamin(adata):
#    artifact = ln.Artifact.from_anndata(adata, description="Test AnnData")
#    gen_config(artifact=artifact)



from __future__ import annotations

import os
import shutil
import tempfile
import time
from pathlib import Path, PurePath
from typing import Literal, get_args
from urllib.parse import urlparse

import requests
from filelock import FileLock
from lamin_utils import logger
from requests.exceptions import RequestException
from rich.progress import Progress

COMPRESSION_FORMATS = Literal["tar.gz", "gztar", "zip", "tar", "gz", "bz", "xz"]
COMPRESSION_FORMATS_LIST = list(get_args(COMPRESSION_FORMATS))
RAW_FORMATS = Literal["csv", "txt", "parquet", "h5ad", "zarr"]
RAW_FORMATS_LIST = list(get_args(RAW_FORMATS))


def _download(
    url: str,
    output_filename: str | None = None,
    output_path: str | Path | None = None,
    archive_format: COMPRESSION_FORMATS | None = None,
    raw_format: RAW_FORMATS | None = None,
    block_size: int = 1024,
    *,
    overwrite: bool = False,
    timeout: int = 30,
    max_retries: int = 3,
    retry_delay: int = 5,
) -> None | Path:  # pragma: no cover
    """Downloads a file irrespective of format.

    Args:
        url: URL to download.
        output_filename: Name of the file to download. If not specified, the file name will be inferred from the URL.
        output_path: Path to download/extract the files to. Defaults to 'ehrapy_data/output_filename' if not specified.
        archive_format: Format of the archive to download. If not specified, the format will be inferred from the URL.
        raw_format: Format of the raw file to download. If not specified, the format will be inferred from the URL.
        block_size: Block size for downloads in bytes.
        overwrite: Whether to overwrite existing files.
        timeout: Request timeout in seconds.
        max_retries: Maximum number of download retries.
        retry_delay: Delay between retries in seconds.
    """

    def _sanitize_filename(filename: str) -> str:
        if os.name == "nt":
            filename = filename.replace("?", "_").replace("*", "_")
        return filename

    def _remove_archive_extension(filename: str) -> str:
        for ext in COMPRESSION_FORMATS_LIST:
            if filename.endswith(ext):
                return filename.removesuffix(ext).rstrip(".")
        return filename

    if output_path is None:
        output_path = tempfile.gettempdir()

    output_path = Path(output_path)

    url_filename = PurePath(urlparse(url).path).name
    suffix = url_filename.split(".")[-1]

    output_filename = _sanitize_filename(url_filename) if output_filename is None else output_filename

    if raw_format is None:
        file_ending = suffix
    elif archive_format is None:
        file_ending = raw_format
    else:
        file_ending = suffix

    if file_ending in RAW_FORMATS_LIST:
        raw_data_output_path = output_path / output_filename
        path_to_check = raw_data_output_path
    elif file_ending in COMPRESSION_FORMATS_LIST:
        tmpdir = tempfile.mkdtemp()
        raw_data_output_path = Path(tmpdir) / output_filename
        path_to_check = output_path / _remove_archive_extension(output_filename)
    else:
        msg = f"Unknown file format: {file_ending}"
        raise RuntimeError(msg)

    lock_path = f"{path_to_check}.lock"
    with FileLock(lock_path, timeout=300):
        if path_to_check.exists():
            warning = f"File {path_to_check} already exists!"
            if not overwrite:
                logger.warning(f"{warning} Using already downloaded dataset...")
                return path_to_check
            else:
                logger.warning(f"{warning} Overwriting...")

        temp_filename = f"{raw_data_output_path}.part"

        retry_count = 0
        while retry_count < max_retries:
            try:
                head_response = requests.head(url, timeout=timeout)
                head_response.raise_for_status()
                content_length = int(head_response.headers.get("content-length", 0))
                free_space = shutil.disk_usage(output_path).free

                if content_length > free_space:
                    msg = f"Insufficient disk space. Need {content_length} bytes, but only {free_space} available."
                    raise OSError(msg)

                response = requests.get(url, stream=True)
                response.raise_for_status()
                total = int(response.headers.get("content-length", 0))

                with Progress(refresh_per_second=5) as progress:
                    task = progress.add_task("[red]Downloading...", total=total)
                    with Path(temp_filename).open("wb") as file:
                        for data in response.iter_content(block_size):
                            file.write(data)
                            progress.update(task, advance=len(data))
                        progress.update(task, completed=total, refresh=True)

                Path(temp_filename).replace(raw_data_output_path)

                if file_ending in COMPRESSION_FORMATS_LIST:
                    shutil.unpack_archive(raw_data_output_path, output_path)

                return path_to_check

            except (OSError, RequestException) as e:
                retry_count += 1
                if retry_count <= max_retries:
                    logger.warning(
                        f"Download attempt {retry_count}/{max_retries} failed: {e!s}. Retrying in {retry_delay} seconds..."
                    )
                    time.sleep(retry_delay)
                else:
                    logger.error(f"Download failed after {max_retries} attempts: {e!s}")
                    if Path(temp_filename).exists():
                        Path(temp_filename).unlink(missing_ok=True)
                    raise

            except Exception as e:
                logger.error(f"Download failed: {e!s}")
                if Path(temp_filename).exists():
                    Path(temp_filename).unlink(missing_ok=True)
                raise
            finally:
                if Path(temp_filename).exists():
                    Path(temp_filename).unlink(missing_ok=True)
                Path(lock_path).unlink(missing_ok=True)

        return path_to_check



import numpy as np
import pandas as pd
import pytest
from tests.conftest import _assert_shape_matches

from ehrdata import EHRData
from ehrdata.io import from_pandas, to_pandas


@pytest.mark.parametrize("df", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
def test_from_pandas_basic(df, request):
    df = request.getfixturevalue(df)
    edata = from_pandas(df)

    _assert_shape_matches(edata, (df.shape[0], df.shape[1], 0), check_R_None=True)
    assert edata.obs.shape == (df.shape[0], 0)
    assert edata.var.shape == (df.shape[1], 0)
    assert edata.tem.shape == (0, 0)

    assert np.array_equal(edata.var_names.values, df.columns.values)

    pd.testing.assert_frame_equal(pd.DataFrame(edata.X), pd.DataFrame(df.to_numpy()))
    desired_default_index = pd.RangeIndex(start=0, stop=len(df), step=1).astype(str)
    assert edata.obs.index.equals(desired_default_index)


@pytest.mark.parametrize("df", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
@pytest.mark.parametrize("index_column", ["patient_id", 0])
def test_from_pandas_basic_index(df, index_column, request):
    df = request.getfixturevalue(df)
    edata = from_pandas(df, index_column=index_column)
    _assert_shape_matches(edata, (df.shape[0], df.shape[1] - 1, 0), check_R_None=True)

    assert edata.obs.shape == (df.shape[0], 0)
    assert edata.var.shape == (df.shape[1] - 1, 0)
    assert edata.tem.shape == (0, 0)

    assert np.array_equal(edata.var_names.values, df.columns.values[1:])

    pd.testing.assert_frame_equal(pd.DataFrame(edata.X), pd.DataFrame(df.iloc[:, 1:].to_numpy()))

    assert np.array_equal(
        edata.obs.index.values,
        df[index_column].values.astype(str)
        if isinstance(index_column, str)
        else df.iloc[:, index_column].values.astype(str),
    )


def test_from_pandas_basic_index_invalid_index_throws_error(csv_basic):
    with pytest.raises(ValueError):
        from_pandas(csv_basic, index_column="mistyped_column_name")

    with pytest.raises(IndexError):
        from_pandas(csv_basic, index_column=6)


@pytest.mark.parametrize("df", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
def test_from_pandas_basic_column_obs_only(df, request):
    df = request.getfixturevalue(df)
    edata = from_pandas(df, columns_obs_only=["patient_id"])
    _assert_shape_matches(edata, (df.shape[0], df.shape[1] - 1, 0), check_R_None=True)
    assert edata.obs.shape == (df.shape[0], 1)
    assert edata.var.shape == (df.shape[1] - 1, 0)
    assert edata.tem.shape == (0, 0)

    assert np.array_equal(edata.var_names.values, df.columns.values[1:])

    pd.testing.assert_frame_equal(pd.DataFrame(edata.X), pd.DataFrame(df.iloc[:, 1:].to_numpy()))
    desired_default_index = pd.RangeIndex(start=0, stop=len(df), step=1).astype(str)
    assert edata.obs.index.equals(desired_default_index)


# avoid using from_pandas here to avoid affecting tests
# of from_pandas and to_pandas
@pytest.mark.parametrize("dataset", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
def test_to_pandas_basic(dataset, request):
    original_df = request.getfixturevalue(dataset)
    edata = EHRData(X=original_df, var=original_df.columns.to_frame())
    df = to_pandas(edata)

    assert df.shape == (original_df.shape[0], original_df.shape[1])
    assert df.index.equals(edata.obs.index)


@pytest.mark.parametrize("dataset", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
def test_to_pandas_basic_layer(dataset, request):
    original_df = request.getfixturevalue(dataset)
    modified_df = original_df.copy()
    modified_df.iloc[1, 1] = 123

    edata = EHRData(X=original_df, var=original_df.columns.to_frame(), layers={"modified_df": modified_df})
    df = to_pandas(edata, layer="modified_df")

    assert df.shape == (original_df.shape[0], original_df.shape[1])
    assert df.index.equals(edata.obs.index)

    assert df.iloc[1, 1] == 123


@pytest.mark.parametrize("dataset", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
@pytest.mark.parametrize("obs_cols", [["patient_id"], {"patient_id"}])
def test_to_pandas_basic_obs_cols(dataset, obs_cols, request):
    original_df = request.getfixturevalue(dataset)
    df_no_patient_id_col = original_df.drop(columns=["patient_id"])

    edata = EHRData(
        X=df_no_patient_id_col, var=df_no_patient_id_col.columns.to_frame(), obs=original_df[["patient_id"]]
    )
    # do this because want to ensure that edata has been set up properly
    assert edata.X.shape == (original_df.shape[0], original_df.shape[1] - 1)
    df = to_pandas(edata, obs_cols=obs_cols)

    assert df.shape == (original_df.shape[0], original_df.shape[1])
    assert df.index.equals(edata.obs.index)

    assert "patient_id" in df.columns


@pytest.mark.parametrize("dataset", ["csv_basic", "csv_non_num_with_missing", "csv_num_with_missing"])
def test_to_pandas_basic_var_col(dataset, request):
    original_df = request.getfixturevalue(dataset)
    edata = EHRData(X=original_df, var=original_df.columns.to_frame())

    new_var_col_name = "new_var_annotation"
    new_var_col_value = [f"some_annotation_{i}" for i in range(edata.shape[1])]
    edata.var[new_var_col_name] = new_var_col_value
    df = to_pandas(edata, var_col=new_var_col_name)

    assert df.shape == original_df.shape
    assert df.index.equals(edata.obs.index)

    assert np.array_equal(df.columns.values, new_var_col_value)
    assert df.shape == (original_df.shape[0], original_df.shape[1])


def test_from_pandas_longitudinal_wide():
    df = pd.DataFrame(
        {
            "var1_t_timestep1": [1, 2, 3],
            "var2_t_timestep1": [4, 5, 6],
            "var1_t_timestep2": [7, 8, 9],
            "var2_t_timestep2": [10, 11, 12],
        },
        index=["patient_1", "patient_2", "patient_3"],
    )
    edata = from_pandas(df, format="wide")
    _assert_shape_matches(edata, (3, 2, 2))

    assert edata.obs.shape == (3, 0)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 0)

    assert np.array_equal(edata.obs_names.values, ["patient_1", "patient_2", "patient_3"])
    assert np.array_equal(edata.var_names.values, ["var1", "var2"])
    assert np.array_equal(edata.tem.index.values, ["timestep1", "timestep2"])

    assert np.array_equal(
        edata.R[1],
        np.array(
            [
                [2, 8],
                [5, 11],
            ]
        ),
    )


def test_from_pandas_longitudinal_wide_missing_timestep():
    df = pd.DataFrame(
        {
            "var1_t_timestep1": [1, 2, 3],
            "var2_t_timestep1": [4, 5, 6],
            "var1_t_timestep2": [7, 8, 9],
            "obs_column": ["obs_1", "obs_2", "obs_3"],
        },
        index=["patient_1", "patient_2", "patient_3"],
    )
    edata = from_pandas(df, format="wide", columns_obs_only=["obs_column"])
    _assert_shape_matches(edata, (3, 2, 2))

    assert edata.obs.shape == (3, 1)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 0)

    assert np.array_equal(edata.obs_names.values, ["patient_1", "patient_2", "patient_3"])
    assert np.array_equal(edata.var_names.values, ["var1", "var2"])
    assert np.array_equal(edata.tem.index.values, ["timestep1", "timestep2"])

    assert np.array_equal(
        edata.R[1],
        np.array(
            [
                [2, 8],
                [5, np.nan],
            ]
        ),
        equal_nan=True,
    )

    assert np.array_equal(edata.obs["obs_column"].values, ["obs_1", "obs_2", "obs_3"])


@pytest.mark.parametrize("columns_obs_only", [["obs_column"], {"obs_column"}])
def test_from_pandas_longitudinal_wide_column_obs_only(columns_obs_only):
    df = pd.DataFrame(
        {
            "var1_t_timestep1": [1, 2, 3],
            "var2_t_timestep1": [4, 5, 6],
            "var1_t_timestep2": [7, 8, 9],
            "var2_t_timestep2": [10, 11, 12],
            "obs_column": ["obs_1", "obs_2", "obs_3"],
        },
        index=["patient_1", "patient_2", "patient_3"],
    )
    edata = from_pandas(df, format="wide", columns_obs_only=columns_obs_only)
    _assert_shape_matches(edata, (3, 2, 2))

    assert edata.obs.shape == (3, 1)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 0)

    assert np.array_equal(edata.obs_names.values, ["patient_1", "patient_2", "patient_3"])
    assert np.array_equal(edata.var_names.values, ["var1", "var2"])
    assert np.array_equal(edata.tem.index.values, ["timestep1", "timestep2"])

    assert np.array_equal(
        edata.R[1],
        np.array(
            [
                [2, 8],
                [5, 11],
            ]
        ),
    )

    assert np.array_equal(edata.obs["obs_column"].values, ["obs_1", "obs_2", "obs_3"])


def test_from_pandas_longitudinal_long():
    df = pd.DataFrame(
        {
            "observation_id": [
                "patient_1",
                "patient_1",
                "patient_1",
                "patient_1",
                "patient_2",
                "patient_2",
                "patient_2",
                "patient_2",
                "patient_3",
                "patient_3",
                "patient_3",
                "patient_3",
            ],
            "variable": [
                "var1",
                "var1",
                "var2",
                "var2",
                "var1",
                "var1",
                "var2",
                "var2",
                "var1",
                "var1",
                "var2",
                "var2",
            ],
            "time": ["t1", "t2", "t1", "t2", "t1", "t2", "t1", "t2", "t1", "t2", "t1", "t2"],
            "value": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        }
    )
    edata = from_pandas(df, format="long")
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.obs.shape == (3, 0)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 0)

    assert np.array_equal(edata.obs_names.values, ["patient_1", "patient_2", "patient_3"])
    assert np.array_equal(edata.var_names.values, ["var1", "var2"])
    assert np.array_equal(edata.tem.index.values, ["t1", "t2"])


def test_from_pandas_invalid_format():
    df = pd.DataFrame()
    with pytest.raises(ValueError):
        from_pandas(df, format="invalid_format")


def test_from_pandas_longitudinal_long_index_column_not_implemented():
    df = pd.DataFrame(
        {
            "observation_id": ["patient_1"],
            "variable": ["var1"],
            "time": ["t1"],
            "value": [1],
        }
    )
    with pytest.raises(ValueError):
        from_pandas(df, index_column="observation_id", format="long")


def test_to_pandas_longitudinal_wide(edata_333):
    df = to_pandas(edata_333, format="wide", layer="R_layer")
    assert df.shape == (3, 9)
    assert df.index.equals(edata_333.obs.index)
    assert np.array_equal(
        df.columns.values,
        [
            "var1_t_t1",
            "var1_t_t2",
            "var1_t_t3",
            "var2_t_t1",
            "var2_t_t2",
            "var2_t_t3",
            "var3_t_t1",
            "var3_t_t2",
            "var3_t_t3",
        ],
    )
    assert np.array_equal(df[["var1_t_t3"]].values.flatten(), edata_333.R[:, 0, 2])


def test_to_pandas_longitudinal_wide_obs_cols(edata_333):
    df = to_pandas(edata_333, format="wide", layer="R_layer", obs_cols=["obs_col_1"])
    assert df.shape == (3, 10)
    assert df.index.equals(edata_333.obs.index)
    assert np.array_equal(
        df.columns.values,
        [
            "var1_t_t1",
            "var1_t_t2",
            "var1_t_t3",
            "var2_t_t1",
            "var2_t_t2",
            "var2_t_t3",
            "var3_t_t1",
            "var3_t_t2",
            "var3_t_t3",
            "obs_col_1",
        ],
    )
    assert np.array_equal(df[["var1_t_t3"]].values.flatten(), edata_333.R[:, 0, 2])
    assert np.array_equal(df[["obs_col_1"]].values.flatten(), edata_333.obs["obs_col_1"].values)


def test_to_pandas_longitudinal_long(edata_333):
    df = to_pandas(edata_333, format="long", layer="R_layer")
    assert df.shape == (27, 4)
    assert np.array_equal(df.iloc[13, :3].values, np.array(["obs2", "var2", "t2"]).astype(object))
    assert df.iloc[13, 3] == 14


def test_to_pandas_invalid_format(edata_333):
    with pytest.raises(ValueError):
        to_pandas(edata_333, format="invalid_format")


def test_to_pandas_longitudinal_long_obs_cols_raises_error(edata_333):
    with pytest.raises(NotImplementedError):
        to_pandas(edata_333, format="long", obs_cols=["obs_col_1"])



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
from dateutil.parser import isoparse  # type: ignore
from fast_array_utils.conv import to_dense
from lamin_utils import logger
from rich import print
from rich.tree import Tree
from scipy.sparse import issparse

from ehrdata.core.constants import CATEGORICAL_TAG, DATE_TAG, FEATURE_TYPE_KEY, MISSING_VALUES, NUMERIC_TAG

if TYPE_CHECKING:
    from collections.abc import Iterable

    from ehrdata import EHRData


def _detect_feature_type(col: pd.Series) -> tuple[Literal["date", "categorical", "numeric"], bool]:
    """Detect the feature type of a :class:`~pandas.Series`.

    Args:
        col: The series to detect the feature type of.

    Returns:
        The detected feature type (one of 'date', 'categorical', or 'numeric') and a boolean, which is True if the feature type is uncertain.
    """
    n_elements = len(col)
    col = col.replace(MISSING_VALUES, np.nan)
    col = col.dropna()
    if len(col) == 0:
        err_msg = f"Feature '{col.name}' contains only NaN values. Please drop this feature to infer the feature type."
        raise ValueError(err_msg)
    majority_type = col.apply(type).value_counts().idxmax()

    if majority_type == pd.Timestamp:
        return DATE_TAG, False  # type: ignore

    if majority_type is str:
        try:
            col.apply(isoparse)
            return DATE_TAG, False  # type: ignore
        except ValueError:
            try:
                col = pd.to_numeric(col, errors="raise")  # Could be an encoded categorical or a numeric feature
                majority_type = float
            except ValueError:
                # Features stored as Strings that cannot be converted to float are assumed to be categorical
                return CATEGORICAL_TAG, False  # type: ignore

    if majority_type not in [int, float]:
        return CATEGORICAL_TAG, False  # type: ignore

    # Guess categorical if the feature is an integer and the values are 0/1 to n-1/n with no gaps
    if (
        (majority_type is int or (np.all(i.is_integer() for i in col)))
        and (n_elements != col.nunique())
        and (
            (col.min() == 0 and np.all(np.sort(col.unique()) == np.arange(col.nunique())))
            or (col.min() == 1 and np.all(np.sort(col.unique()) == np.arange(1, col.nunique() + 1)))
        )
    ):
        return CATEGORICAL_TAG, True  # type: ignore

    return NUMERIC_TAG, False  # type: ignore


def infer_feature_types(
    edata: EHRData,
    *,
    layer: str | None = None,
    output: Literal["tree", "dataframe"] | None = "tree",
    verbose: bool = True,
) -> pd.DataFrame | None:
    """Infer feature types from an :class:`~ehrdata.EHRData` object.

    For each feature in `edata.var_names`, the method infers one of the following types: `'date'`, `'categorical'`, or `'numeric'`.
    The inferred types are stored in `edata.var['feature_type']`. Please check the inferred types and adjust if necessary using
    `edata.var['feature_type']['feature1']='corrected_type'` or with :func:`~ehrdata.replace_feature_types`.
    Be aware that not all features stored numerically are of `'numeric'` type, as categorical features might be stored in a numerically encoded format.
    For example, a feature with values [0, 1, 2] might be a categorical feature with three categories. This is accounted for in the method, but it is
    recommended to check the inferred types.

    Args:
        edata: Data object.
        layer: The layer to use from the EHRData object. If `None`, the `X` field is used.
        output: The output format. Choose between `'tree'`, `'dataframe'`, or `None`. If `'tree'`, the feature types will be printed to the console in a tree format.
            If `'dataframe'`, a :class:`~pandas.DataFrame` with the feature types will be returned. If `None`, nothing will be returned.
        verbose: Whether to print warnings for uncertain feature types.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.mimic_2()
        >>> ed.infer_feature_types(edata)
    """
    feature_types = {}
    uncertain_features = []

    X = edata.X if layer is None else edata.layers[layer]

    if issparse(X):
        X = to_dense(X)

    df = pd.DataFrame(X.reshape(-1, edata.shape[1]), columns=edata.var_names)

    for feature in edata.var_names:
        if (
            FEATURE_TYPE_KEY in edata.var
            and edata.var[FEATURE_TYPE_KEY][feature] is not None
            and not pd.isna(edata.var[FEATURE_TYPE_KEY][feature])
        ):
            feature_types[feature] = edata.var[FEATURE_TYPE_KEY][feature]
        else:
            feature_types[feature], raise_warning = _detect_feature_type(df[feature])
            if raise_warning:
                uncertain_features.append(feature)

    edata.var[FEATURE_TYPE_KEY] = pd.Series(feature_types)[edata.var_names]

    if verbose:
        logger.warning(
            f"{'Features' if len(uncertain_features) > 1 else 'Feature'} {str(uncertain_features)[1:-1]} {'were' if len(uncertain_features) > 1 else 'was'} detected as categorical features stored numerically."
            f"Please verify and adjust if necessary using `ed.replace_feature_types`."
        )

        logger.info(
            f"Stored feature types in edata.var['{FEATURE_TYPE_KEY}']."
            f" Please verify and adjust if necessary using `ed.replace_feature_types`."
        )

    if output == "tree":
        feature_type_overview(edata)
    elif output == "dataframe":
        return edata.var[FEATURE_TYPE_KEY].to_frame()
    elif output is not None:
        err_msg = f"Output format {output} not recognized. Choose between 'tree', 'dataframe', or None."
        raise ValueError(err_msg)


# TODO: this function is a different flavor of inferring feature types. We should decide on a single implementation in the future.
def _infer_numerical_column_indices(
    edata: EHRData, layer: str | None = None, column_indices: Iterable[int] | None = None
) -> list[int]:
    mtx = edata.X if layer is None else edata[layer]
    indices = (
        list(range(mtx.shape[1])) if column_indices is None else [i for i in column_indices if i < mtx.shape[1] - 1]
    )
    non_numerical_indices = []
    for i in indices:
        # The astype("float64") call will throw only if the feature's data type cannot be cast to float64, meaning in
        # practice it contains non-numeric values. Consequently, it won't throw if the values are numeric but stored
        # as an "object" dtype, as astype("float64") can successfully convert them to floats.
        try:
            mtx[::, i].astype("float64")
        except ValueError:
            non_numerical_indices.append(i)

    return [idx for idx in indices if idx not in non_numerical_indices]


def _check_feature_types(func):
    @wraps(func)
    def wrapper(edata, *args, **kwargs):
        from ehrdata import EHRData

        # Account for class methods that pass self as first argument
        _self = None
        if not isinstance(edata, EHRData) and len(args) > 0 and isinstance(args[0], EHRData):
            _self = edata
            edata = args[0]
            args = args[1:]

        if FEATURE_TYPE_KEY not in edata.var:
            infer_feature_types(edata, output=None)
            logger.warning(
                f"Feature types were inferred and stored in edata.var[{FEATURE_TYPE_KEY}]. Please verify using `ed.feature_type_overview` and adjust if necessary using `ed.replace_feature_types`."
            )

        for feature in edata.var_names:
            feature_type = edata.var[FEATURE_TYPE_KEY][feature]
            if (
                feature_type is not None
                and (not pd.isna(feature_type))
                and feature_type not in [CATEGORICAL_TAG, NUMERIC_TAG, DATE_TAG]
            ):
                logger.warning(
                    f"Feature '{feature}' has an invalid feature type '{feature_type}'. Please correct using `ed.replace_feature_types`."
                )

        if _self is not None:
            return func(_self, edata, *args, **kwargs)
        return func(edata, *args, **kwargs)

    return wrapper


@_check_feature_types
def feature_type_overview(edata: EHRData) -> None:
    """Print an overview of the feature types and encoding modes in the :class:`~ehrdata.EHRData` object.

    Args:
        edata: Data object.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.mimic_2()
        >>> ed.feature_type_overview(edata)
    """
    tree = Tree(
        f"[b] Detected feature types for EHRData object with {len(edata.obs_names)} obs and {len(edata.var_names)} vars",
        guide_style="underline2",
    )

    branch = tree.add("📅[b] Date features")
    for date in sorted(edata.var_names[edata.var[FEATURE_TYPE_KEY] == DATE_TAG]):
        branch.add(date)

    branch = tree.add("📐[b] Numerical features")
    for numeric in sorted(edata.var_names[edata.var[FEATURE_TYPE_KEY] == NUMERIC_TAG]):
        branch.add(numeric)

    branch = tree.add("🗂️[b] Categorical features")
    cat_features = edata.var_names[edata.var[FEATURE_TYPE_KEY] == CATEGORICAL_TAG]

    df = pd.DataFrame(
        to_dense(edata[:, cat_features].X) if issparse(edata[:, cat_features].X) else edata[:, cat_features].X,
        columns=cat_features,
    )

    if "encoding_mode" in edata.var:
        unencoded_vars = edata.var.loc[cat_features, "unencoded_var_names"].unique().tolist()

        for unencoded in sorted(unencoded_vars):
            if unencoded in edata.var_names:
                branch.add(f"{unencoded} ({df.loc[:, unencoded].nunique()} categories)")
            else:
                enc_mode = edata.var.loc[edata.var["unencoded_var_names"] == unencoded, "encoding_mode"].values[0]
                branch.add(f"{unencoded} ({edata.obs[unencoded].nunique()} categories); {enc_mode} encoded")

    else:
        for categorical in sorted(cat_features):
            categorical_feature = df.loc[:, categorical]
            categorical_feature_nans_cleaned = categorical_feature.replace(MISSING_VALUES, np.nan)
            branch.add(f"{categorical} ({categorical_feature_nans_cleaned.nunique()} categories)")

    print(tree)


def replace_feature_types(
    edata: EHRData,
    features: Iterable[str],
    corrected_type: Literal["categorical", "numeric", "date"],
) -> None:
    """Correct the feature types for a list of features inplace.

    Args:
        edata: Data object.
        features: The features to correct.
        corrected_type: The corrected feature type. One of `'date'`, `'categorical'`, or `'numeric'`.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.diabetes_130_fairlearn()
        >>> ed.infer_feature_types(edata)
        >>> ed.replace_feature_types(edata, ["time_in_hospital", "number_diagnoses", "num_procedures"], "numeric")
    """
    if FEATURE_TYPE_KEY not in edata.var:
        err_msg = "Feature types were not inferred. Please infer feature types using 'ed.infer_feature_types' before correcting."
        raise ValueError(err_msg)

    if corrected_type not in [CATEGORICAL_TAG, NUMERIC_TAG, DATE_TAG]:
        err_msg = f"Corrected type {corrected_type} not recognized. Choose between '{DATE_TAG}', '{CATEGORICAL_TAG}', or '{NUMERIC_TAG}'."
        raise ValueError(err_msg)

    if FEATURE_TYPE_KEY not in edata.var:
        err_msg = (
            "Feature types were not inferred. Please infer feature types using 'infer_feature_types' before correcting."
        )
        raise ValueError(err_msg)

    if isinstance(features, str):
        features = [features]

    edata.var.loc[features, FEATURE_TYPE_KEY] = corrected_type


def harmonize_missing_values(
    edata: EHRData,
    *,
    layer: str | None = None,
    missing_values: Iterable[str] | None = ["nan", "np.nan", "<NA>", "pd.NA"],
    copy: bool = False,
) -> EHRData | None:
    """Harmonize missing values in the :class:`~ehrdata.EHRData` object.

    This function will replace strings that are considered to represent missing values with `np.nan`.

    Args:
        edata: Data object.
        layer: The layer to use from the :class:`~ehrdata.EHRData` object. If `None`, the `X` layer is used.
        missing_values: The strings that are considered to represent missing values and should be replaced with np.nan
        copy: Whether to return a copy of the :class:`~ehrdata.EHRData` object with the missing values replaced.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.mimic_2()
        >>> ed.harmonize_missing_values(edata)
    """
    if copy:
        edata = edata.copy()
    X = edata.X if layer is None else edata.layers[layer]

    # note that every sparse array is of a numeric dtype and will enter this if block
    if np.issubdtype(X.dtype, np.number):
        logger.warning(f"This operation does not affect numeric layer {'X' if layer is None else layer}.")
        return edata if copy else None

    df = pd.DataFrame(X.reshape(-1, edata.shape[1]), columns=edata.var_names)
    df.replace(missing_values, np.nan, inplace=True)

    if layer is None:
        edata.X = df.values.reshape(X.shape)
    else:
        edata.layers[layer] = df.values.reshape(X.shape)

    return edata if copy else None



def test_repr(edata_333):
    edata = edata_333

    # vanilla case
    repr_str = edata.__repr__()
    target_repr_str = "EHRData object with n_obs × n_vars × n_t = 3 × 3 × 3\n    obs: 'obs_col_1'\n    var: 'var_col_1'\n    tem: 't1', 't2', 't3'\n    shape of .X: (3, 3)\n    shape of .R: (3, 3, 3)"
    assert repr_str == target_repr_str

    # more layers
    edata.layers["test_layer"] = edata.X.copy()
    repr_str_with_layer = edata.__repr__()
    target_repr_str_with_layer = "EHRData object with n_obs × n_vars × n_t = 3 × 3 × 3\n    obs: 'obs_col_1'\n    var: 'var_col_1'\n    tem: 't1', 't2', 't3'\n    layers: 'test_layer'\n    shape of .X: (3, 3)\n    shape of .R: (3, 3, 3)"
    assert repr_str_with_layer == target_repr_str_with_layer

    # view
    edata_view = edata[:2, :2, :2]
    repr_str_view = edata_view.__repr__()
    target_repr_str_view = "View of EHRData object with n_obs × n_vars × n_t = 2 × 2 × 2\n    obs: 'obs_col_1'\n    var: 'var_col_1'\n    tem: 't1', 't2'\n    layers: 'test_layer'\n    shape of .X: (2, 2)\n    shape of .R: (2, 2, 2)"
    assert repr_str_view == target_repr_str_view



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
import xarray as xr
from fast_array_utils.conv import to_dense
from scipy.sparse import issparse

if TYPE_CHECKING:
    from collections.abc import Iterable

    from ehrdata import EHRData

from ehrdata.core.constants import PANDAS_FORMATS


def from_pandas(
    df: pd.DataFrame,
    *,
    columns_obs_only: Iterable[str] | None = None,
    index_column: str | int | None = None,
    format: Literal["flat", "wide", "long"] = "flat",
    wide_format_time_suffix: str | None = None,
    long_format_keys: dict[Literal["observation_column", "variable_column", "time_column", "value_column"], str]
    | None = None,
) -> EHRData:
    """Transform a given :class:`~pandas.DataFrame` into an :class:`~ehrdata.EHRData` object.

    Note that columns containing boolean values (either 0/1 or T(t)rue/F(f)alse)
    will be stored as boolean columns.
    The other non-numerical columns will be stored as categorical values.

    Args:
        df: The dataframe to be transformed.
        columns_obs_only: Column names that should belong to `obs` only and not `X`.
        index_column: The index column of `obs`.
            This can be either a column name (or its numerical index in the DataFrame) or the index of the dataframe.
        format: The format of the input dataframe.
            If the data is not longitudinal, choose `format="flat"`.
            If the data is longitudinal in the long format, choose `format="long"`.
            If the data is longitudinal in a wide format, choose `format="wide"`.
        wide_format_time_suffix: Use only if `format="wide"`.
            Suffices in the variable columns that indicate the time of the observation.
            The collected suffices will be sorted lexicographically.
            The variables will be ordered accordingly along the 3rd axis of the :class:`~ehrdata.EHRData` object.
        long_format_keys: Use only if `format="long"`.
            The keys of the dataframe in the long format.
            The dictionary should have the following structure: `{"observation_column": "<the column name of the observation ids>",
            "variable_column": "<the column name of the variable ids>",
            "time_column": "<the column name of the time>",
            "value_column": "<the column name of the values>"}`.

    Examples:
        >>> import ehrdata as ed
        >>> import pandas as pd
        >>> df = pd.DataFrame(
        ...     {
        ...         "patient_id": ["0", "1", "2", "3", "4"],
        ...         "age": [65, 72, 58, 78, 82],
        ...         "sex": ["M", "F", "F", "M", "F"],
        ...     }
        ... )
        >>> edata = ed.io.from_pandas(df, index_column="patient_id")
        >>> edata

        >>> EHRData object with n_obs × n_vars = 5 × 2
        >>> shape of .X: (5, 2)

        >>> df_wide = pd.DataFrame(
        ...     {
        ...         "patient_id": ["0", "1"],
        ...         "sex": ["F", "M"],
        ...         "systolic_blood_pressure_t_0": [120, 130],
        ...         "systolic_blood_pressure_t_2": [125, np.nan],  # the suffix strings are sorted lexicographically
        ...         "systolic_blood_pressure_t_1": [np.nan, 135],
        ...     }
        ... )
        >>> edata = ed.io.from_pandas(df_wide, format="wide", columns_obs_only=["patient_id", "sex"])
        >>> edata

        >>> EHRData object with n_obs × n_vars × n_t = 2 × 1 × 3
        >>> obs: 'patient_id', 'sex'
        >>> shape of .X: (2, 1)
        >>> shape of .R: (2, 1, 3)


        >>> df_long = pd.DataFrame(
        ...     {
        ...         "observation_id": ["0", "0", "0", "1", "1", "1"],
        ...         "variable": [
        ...             "sex",
        ...             "systolic_blood_pressure",
        ...             "systolic_blood_pressure",
        ...             "sex",
        ...             "systolic_blood_pressure",
        ...             "systolic_blood_pressure",
        ...         ],
        ...         "time": ["t_0", "t_0", "t_1", "t_0", "t_0", "t_2"],
        ...         "value": ["F", 120, 125, "M", 130, 135],
        ...     }
        ... )
        >>> edata = ed.io.from_pandas(df_long, format="long", columns_obs_only=["sex"])
        >>> edata

        >>> EHRData object with n_obs × n_vars × n_t = 2 × 1 × 3
        >>> obs: "sex"
        >>> shape of .X: (2, 1)
        >>> shape of .R: (2, 1, 3)

    """
    from ehrdata import EHRData

    if format not in PANDAS_FORMATS:
        err_msg = f"Format {format} is not supported. Please choose from {PANDAS_FORMATS}."
        raise ValueError(err_msg)

    if format != "long" and long_format_keys is not None:
        err_msg = "long_format_keys should only be used if format is 'long'."
        raise ValueError(err_msg)

    elif format == "long" and long_format_keys is None:
        long_format_keys = {
            "observation_column": "observation_id",
            "variable_column": "variable",
            "time_column": "time",
            "value_column": "value",
        }
    elif format == "long" and long_format_keys is not None:
        valid_long_format_keys = ["observation_column", "variable_column", "time_column", "value_column"]
        invalid_keys = [key for key in long_format_keys if key not in valid_long_format_keys]
        if invalid_keys:
            err_msg = f"Invalid keys: {invalid_keys}. Please use only the following keys: {valid_long_format_keys}."
            raise ValueError(err_msg)

    if format != "wide" and wide_format_time_suffix is not None:
        err_msg = "wide_format_time_suffix should only be used if format is 'wide'."
        raise ValueError(err_msg)
    elif format == "wide" and wide_format_time_suffix is None:
        wide_format_time_suffix = "_t_"

    # index column:
    if format in ["flat", "wide"]:
        if index_column is not None:
            # Because an integer is allowed to specify the index column, we need to check if it is out of bounds.
            if isinstance(index_column, int):
                if index_column >= len(df.columns):
                    err_msg = f"index_column integer index is out of bounds. DataFrame has {len(df.columns)} columns."
                    raise IndexError(err_msg)
                index_column = df.columns[index_column]
            # Because the column name is allowed to specify the index column, we need to check it is a valid column name.
            if not df.index.name or df.index.name != index_column:
                if index_column in df.columns:
                    df = df.set_index(index_column)
                else:
                    err_msg = f"Column {index_column} not found in DataFrame."
                    raise ValueError(err_msg)

        # Now handle columns_obs_only with consideration of the new index
        if columns_obs_only:
            columns_obs_only = list(columns_obs_only)
            if index_column in columns_obs_only:
                columns_obs_only.remove(index_column)
            missing_cols = [col for col in columns_obs_only if col not in df.columns]
            if missing_cols:
                err_msg = f"Columns {missing_cols} specified in columns_obs_only are not in the DataFrame."
                raise ValueError(err_msg)
            obs = df.loc[:, columns_obs_only].copy()
            df = df.drop(columns=columns_obs_only, errors="ignore")
        else:
            obs = pd.DataFrame(index=df.index)

        for col in obs.columns:
            if obs[col].dtype == "bool":
                obs[col] = obs[col].astype(bool)
            elif obs[col].dtype == "object":
                obs[col] = obs[col].astype("category")

    if format == "flat":
        X = df.to_numpy()
        obs.index = obs.index.astype(str)
        var = pd.DataFrame(index=df.columns)
        var.index = var.index.astype(str)

        # Handle dtype of X based on presence of numerical columns only
        all_numeric = df.select_dtypes(include=[np.number]).shape[1] == df.shape[1]
        X = X.astype(np.float64 if all_numeric else object)

        edata = EHRData(X=X, obs=obs, var=var)

    elif format == "wide":
        # Define the variable name without suffix for each column.
        try:
            variables, timepoints = zip(*[col.split(wide_format_time_suffix) for col in df.columns], strict=False)
        except ValueError:
            err_msg = f"The wide format time suffix {wide_format_time_suffix} is not found in the column names. For variables not containing the time suffix, please put them as static variables into `obs` using the `columns_obs_only` parameter. Please specify the time suffix using the wide_format_time_suffix parameter."
            raise ValueError(err_msg) from None

        unique_variables = np.array(sorted(set(variables)))
        unique_timepoints = np.array(sorted(set(timepoints)))
        variables = np.array(variables)
        timepoints = np.array(timepoints)

        R = np.full((len(df), len(unique_variables), len(unique_timepoints)), np.nan)
        for i, timepoint in enumerate(unique_timepoints):
            for j, variable in enumerate(unique_variables):
                if variable + wide_format_time_suffix + timepoint in df.columns:
                    R[:, j, i] = df.loc[:, (variables == variable) & (timepoints == timepoint)].to_numpy().flatten()

        obs.index = obs.index.astype(str)
        var = pd.DataFrame(index=unique_variables)
        var.index = var.index.astype(str)
        tem = pd.DataFrame(index=unique_timepoints)
        tem.index = tem.index.astype(str)

        edata = EHRData(R=R, obs=obs, var=var, tem=tem)

    elif format == "long":
        if index_column is not None:
            err_msg = (
                "For long format, use the `observation_column` parameter to specify the index of the observations."
            )
            raise ValueError(err_msg)

        unique_observations = df[long_format_keys["observation_column"]].unique()
        obs = pd.DataFrame(index=unique_observations)

        if columns_obs_only is not None:
            df_obs_only = df[df[long_format_keys["variable_column"]].isin(columns_obs_only)].drop(
                columns=long_format_keys["time_column"]
            )

            df_obs_pivoted = pd.pivot(
                df_obs_only, index=long_format_keys["observation_column"], columns=long_format_keys["variable_column"]
            )
            df_obs_pivoted.columns = df_obs_pivoted.columns.droplevel(0)

            obs = obs.join(df_obs_pivoted, how="left")

        if columns_obs_only is not None:
            df = df[~df[long_format_keys["variable_column"]].isin(columns_obs_only)]
        xr_dataarray = df.set_index(
            [
                long_format_keys["observation_column"],
                long_format_keys["variable_column"],
                long_format_keys["time_column"],
            ]
        ).to_xarray()

        R = xr_dataarray[long_format_keys["value_column"]].values

        var = pd.DataFrame(index=xr_dataarray[long_format_keys["variable_column"]].values)
        var.index = var.index.astype(str)
        tem = pd.DataFrame(index=xr_dataarray[long_format_keys["time_column"]].values)
        tem.index = tem.index.astype(str)

        edata = EHRData(R=R, obs=obs, var=var, tem=tem)

    edata.obs_names = edata.obs_names.astype(str)
    edata.var_names = edata.var_names.astype(str)

    return edata


def to_pandas(
    edata: EHRData,
    *,
    layer: str | None = None,
    obs_cols: Iterable[str] | None = None,
    var_col: str | None = None,
    format: Literal["wide", "long"] = "wide",
) -> pd.DataFrame:
    """Transform an :class:`~ehrdata.EHRData` object to a :class:`~pandas.DataFrame`.

    Args:
        edata: Data object.
        layer: The layer to access the values of. If not specified, it uses the `X` matrix.
        obs_cols: The columns of `obs` to add to the dataframe.
        var_col: The column of `var` to create the column names from in the created dataframe.
            If not specified, the `var_names` will be used.
        format: The format of the output dataframe.
            This is relevant for longitudinal data.
            If `"wide"`, the output dataframe will write a column for each (variable, time) tuple, naming the column as `<variable_name>_t_<tem.index value>`.
            If `"long"`, the output dataframe will be in long format, with columns `"observation_id"`, `"variable"`, `"time"`, and `"value"`.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.ehrdata_blobs(n_observations=2, n_variables=2, base_timepoints=3)
        >>> edata

        >>> EHRData object with n_obs × n_vars × n_t = 2 × 2 × 3
        >>> obs: "cluster"
        >>> tem: '0', '1', '2'
        >>> shape of .X: (2, 2)
        >>> shape of .R: (2, 2, 3)

        >>> df_wide = ed.io.to_pandas(edata, format="wide")
        >>> df_wide

        +-----+----------------+----------------+----------------+----------------+----------------+----------------+
        |     | feature_0_t_0  | feature_0_t_1  | feature_0_t_2  | feature_1_t_0  | feature_1_t_1  | feature_1_t_2  |
        +=====+================+================+================+================+================+================+
        |  0  |    3.060372    |    3.827524    |    4.680650    |   -1.697623    |   -1.816282    |   -2.775774    |
        +-----+----------------+----------------+----------------+----------------+----------------+----------------+
        |  1  |   -3.395852    |   -4.948999    |   -5.401154    |   -7.347151    |   -9.427101    |  -11.793235    |
        +-----+----------------+----------------+----------------+----------------+----------------+----------------+

        >>> df_long = ed.io.to_pandas(edata, format="long")
        >>> df_long

        +--------------------+------------------+------+-------------+
        | observation_id     | variable         | time | value       |
        +====================+==================+======+=============+
        | 0                  | feature_0        | 0    | 3.060372    |
        +--------------------+------------------+------+-------------+
        | 0                  | feature_0        | 1    | 3.827524    |
        +--------------------+------------------+------+-------------+
        | 0                  | feature_0        | 2    | 4.680650    |
        +--------------------+------------------+------+-------------+
        | 0                  | feature_1        | 0    | -1.697623   |
        +--------------------+------------------+------+-------------+
        | 0                  | feature_1        | 1    | -1.816282   |
        +--------------------+------------------+------+-------------+
        | 0                  | feature_1        | 2    | -2.775774   |
        +--------------------+------------------+------+-------------+
        | 1                  | feature_0        | 0    | -3.395852   |
        +--------------------+------------------+------+-------------+
        | 1                  | feature_0        | 1    | -4.948999   |
        +--------------------+------------------+------+-------------+
        | 1                  | feature_0        | 2    | -5.401154   |
        +--------------------+------------------+------+-------------+
        | 1                  | feature_1        | 0    | -7.347151   |
        +--------------------+------------------+------+-------------+
        | 1                  | feature_1        | 1    | -9.427101   |
        +--------------------+------------------+------+-------------+
        | 1                  | feature_1        | 2    | -11.793235  |
        +--------------------+------------------+------+-------------+
    """
    if format not in PANDAS_FORMATS:
        err_msg = f"Format {format} is not supported. Please choose from {PANDAS_FORMATS}."
        raise ValueError(err_msg)

    if var_col is not None and var_col not in edata.var.columns:
        err_msg = f"Variable column {var_col} not found in edata.var"
        raise ValueError(err_msg)

    if layer is not None:
        if layer == "X":
            layer = None
        if layer == "R":
            layer = "R_layer"
    X = edata.layers[layer] if layer is not None else edata.X

    var_names = edata.var_names if var_col is None else edata.var[var_col]

    if issparse(X):  # pragma: no cover
        X = to_dense(X)

    if format == "wide":
        if len(X.shape) == 2:
            df = pd.DataFrame(X, columns=var_names)
        elif len(X.shape) == 3:
            X_wide = X.reshape(X.shape[0], -1)
            column_names = [
                f"{edata.var_names[i]}_t_{edata.tem.index[j]}" for i in range(X.shape[1]) for j in range(X.shape[2])
            ]
            df = pd.DataFrame(X_wide, columns=column_names)

        if obs_cols:
            obs_cols = list(obs_cols)
            if len(edata.obs.columns) == 0:
                msg = "Cannot slice columns from empty obs!"
                raise ValueError(msg)
            obs_slice = edata.obs[obs_cols]
            # reset index needed since we slice all or at least some columns from obs DataFrame
            obs_slice = obs_slice.reset_index(drop=True)
            df = pd.concat([df, obs_slice], axis=1)
        df.index = edata.obs_names
        return df

    elif format == "long":
        if obs_cols:
            err_msg = "Long format does not support obs_cols"
            raise NotImplementedError(err_msg)

        if len(X.shape) == 2:
            df = pd.DataFrame(X, columns=var_names)
            df = df.melt(id_vars=edata.obs_names, var_name="variable", value_name="value")
            # to long
        elif len(X.shape) == 3:
            data_array = xr.DataArray(
                X,
                dims=["observation_id", "variable", "time"],
                coords={
                    "observation_id": edata.obs_names,
                    "variable": edata.var_names,
                    "time": edata.tem.index,
                },
            )
            data_array.name = "value"
            df = data_array.to_dataframe().reset_index()

        return df



import pandas as pd
import pytest
from tests.conftest import TEST_DATA_PATH, _assert_shape_matches

from ehrdata.io import read_csv

TEST_PATH = TEST_DATA_PATH / "toy_csv"


@pytest.mark.parametrize(
    ("filename", "target_shape"),
    [
        ("csv_basic.csv", (5, 4)),
        ("csv_non_num_with_missing.csv", (5, 6)),
        ("csv_num_with_missing.csv", (5, 4)),
    ],
)
def test_read_csv(filename, target_shape):
    edata = read_csv(TEST_PATH / filename)

    _assert_shape_matches(edata, (target_shape[0], target_shape[1], 0), check_R_None=True)

    assert edata.obs.shape == (target_shape[0], 0)
    assert edata.var.shape == (target_shape[1], 0)
    assert edata.tem.shape == (0, 0)

    desired_default_index = pd.RangeIndex(start=0, stop=target_shape[0], step=1).astype(str)
    assert edata.obs.index.equals(desired_default_index)


@pytest.mark.parametrize(
    ("filename", "target_shape"),
    [
        ("csv_basic.csv", (5, 4)),
        ("csv_non_num_with_missing.csv", (5, 6)),
        ("csv_num_with_missing.csv", (5, 4)),
    ],
)
@pytest.mark.parametrize("index_column", ["patient_id", 0])
def test_read_csv_index_column(filename, target_shape, index_column):
    edata = read_csv(TEST_PATH / filename, index_column=index_column)

    _assert_shape_matches(edata, (target_shape[0], target_shape[1] - 1, 0), check_R_None=True)
    assert edata.obs.shape == (target_shape[0], 0)
    assert edata.var.shape == (target_shape[1] - 1, 0)
    assert edata.tem.shape == (0, 0)

    assert edata.obs.index.name == "patient_id"


@pytest.mark.parametrize(
    ("filename", "target_shape"),
    [
        ("csv_basic.csv", (5, 4)),
        ("csv_non_num_with_missing.csv", (5, 6)),
        ("csv_num_with_missing.csv", (5, 4)),
    ],
)
def test_read_csv_columns_obs_only(filename, target_shape):
    edata = read_csv(TEST_PATH / filename, columns_obs_only=["patient_id"])

    _assert_shape_matches(edata, (target_shape[0], target_shape[1] - 1, 0), check_R_None=True)
    assert edata.obs.shape == (target_shape[0], 1)
    assert edata.var.shape == (target_shape[1] - 1, 0)
    assert edata.tem.shape == (0, 0)

    desired_default_index = pd.RangeIndex(start=0, stop=target_shape[0], step=1).astype(str)
    assert edata.obs.index.equals(desired_default_index)

    assert "patient_id" in edata.obs.columns



from pathlib import Path

R_LAYER_KEY = "R_layer"


# Typing Column
# -----------------------
# The column name and used values in adata.var for column types.

FEATURE_TYPE_KEY = "feature_type"
NUMERIC_TAG = "numeric"
CATEGORICAL_TAG = "categorical"
DATE_TAG = "date"


# Missing values
# --------------
# These values if encountered as strings are considered to represent missing values in the data

MISSING_VALUES = (
    "nan",  # this is the result value of str(np.nan)
    "np.nan",  # this is very explicit about what the value should mean
    "<NA>",  # this is the result value of str(pd.NA)
    "pd.NA",  # this is very explicit about what the value should mean
)


# Pandas Format
# --------------
# The format of the pandas dataframe to be used in the to_pandas and from_pandas functions

PANDAS_FORMATS = ["flat", "wide", "long"]


# Default data path
# ------------------
# The default path to store and load data

DEFAULT_DATA_PATH = Path("ehrapy_data")



import anndata as ad
import numpy as np
import pandas as pd
import pytest
from tests.conftest import _assert_shape_matches

from ehrdata import EHRData
from ehrdata.core.constants import R_LAYER_KEY


def _assert_fields_are_view(edata: EHRData):
    assert edata.is_view
    assert isinstance(edata.X, ad._core.views.ArrayView)
    assert isinstance(edata.R, ad._core.views.ArrayView)
    assert isinstance(edata.tem, ad._core.views.DataFrameView)


#################################################################
### Test combinations of X, R, t during initialization
#################################################################
def test_ehrdata_init_vanilla_empty():
    edata = EHRData()

    _assert_shape_matches(edata, (0, 0, 0), check_X_None=True, check_R_None=True)
    assert edata.X is None
    assert edata.R is None
    assert edata.obs.shape == (0, 0)
    assert edata.var.shape == (0, 0)
    assert edata.tem.shape == (0, 0)


def test_ehrdata_init_vanilla_X(X_numpy_32):
    edata = EHRData(X=X_numpy_32)
    _assert_shape_matches(edata, (3, 2, 0), check_R_None=True)

    assert edata.obs.shape == (3, 0)

    assert edata.var.shape == (2, 0)

    assert edata.tem.shape == (0, 0)


def test_ehrdata_init_vanilla_r(R_numpy_322):
    edata = EHRData(R=R_numpy_322)
    _assert_shape_matches(edata, (3, 2, 2))

    assert edata.layers is not None
    assert edata.layers[R_LAYER_KEY] is not None
    assert edata.layers[R_LAYER_KEY].shape == (3, 2, 2)

    assert edata.obs.shape == (3, 0)

    assert edata.var.shape == (2, 0)

    assert edata.tem.shape == (2, 0)


def test_ehrdata_init_vanilla_X_and_r(X_numpy_32, R_numpy_322):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322)
    _assert_shape_matches(edata, (3, 2, 2))

    assert edata.layers is not None
    assert edata.R is not None

    assert edata.X.shape == (3, 2)
    assert edata.obs.shape == (3, 0)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 0)
    assert edata.R.shape == (3, 2, 2)


def test_ehrdata_init_vanilla_X_and_t(X_numpy_32, tem_21):
    edata = EHRData(X=X_numpy_32, tem=tem_21)
    _assert_shape_matches(edata, (3, 2, 2), check_R_None=True)

    assert edata.layers is not None

    assert edata.X.shape == (3, 2)
    assert edata.obs.shape == (3, 0)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 1)


def test_ehrdata_init_vanilla_X_and_r_and_t(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322, tem=tem_21)
    _assert_shape_matches(edata, (3, 2, 2))

    assert edata.layers is not None
    assert edata.R is not None

    assert edata.X.shape == (3, 2)
    assert edata.obs.shape == (3, 0)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (2, 1)
    assert edata.R.shape == (3, 2, 2)


def test_ehrdata_init_vanilla_X_and_layers(X_numpy_32):
    edata = EHRData(X=X_numpy_32, layers={"some_layer": X_numpy_32})

    _assert_shape_matches(edata, (3, 2, 0), check_R_None=True)

    assert edata.obs.shape == (3, 0)
    assert edata.var.shape == (2, 0)
    assert edata.tem.shape == (0, 0)


#################################################################
### Test aligned dataframes during intitialization
#################################################################
def test_ehrdata_init_vanilla_obs(obs_31):
    edata = EHRData(obs=obs_31)
    _assert_shape_matches(edata, (3, 0, 0), check_X_None=True, check_R_None=True)


def test_ehrdata_init_vanilla_var(var_31):
    edata = EHRData(var=var_31)
    _assert_shape_matches(edata, (0, 3, 0), check_X_None=True, check_R_None=True)


def test_ehrdata_init_vanilla_tem(tem_31):
    edata = EHRData(tem=tem_31)
    _assert_shape_matches(edata, (0, 0, 3), check_X_None=True, check_R_None=True)


#################################################################
### Test assignment of X, R, t combinations
#################################################################
def test_ehrdata_init_r_assign_X_tem(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(R=R_numpy_322)
    edata.X = X_numpy_32
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 0)

    edata = EHRData(R=R_numpy_322)
    edata.tem = tem_21
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 1)

    # for assignment of X and t, test both orders
    edata = EHRData(R=R_numpy_322)
    edata.X = X_numpy_32
    edata.tem = tem_21
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 1)

    edata = EHRData(R=R_numpy_322)
    edata.tem = tem_21
    edata.X = X_numpy_32
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 1)


def test_ehrdata_init_X_assign_r_t(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(X=X_numpy_32)
    with pytest.raises(ValueError):
        edata.R = R_numpy_322

    edata = EHRData(X=X_numpy_32)
    with pytest.raises(ValueError):
        edata.tem = tem_21


def test_ehrdata_init_t_assign_X_r(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(tem=tem_21)
    with pytest.raises(ValueError):
        edata.X = X_numpy_32

    edata = EHRData(tem=tem_21)
    with pytest.raises(ValueError):
        edata.R = R_numpy_322


def test_ehrdata_init_X_r_assign_t(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322)

    edata.tem = tem_21
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 1)


def test_ehrdata_init_X_t_assign_r(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(X=X_numpy_32, tem=tem_21)

    edata.R = R_numpy_322
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 1)


def test_ehrdata_init_r_t_assign_X(X_numpy_32, R_numpy_322, tem_21):
    edata = EHRData(R=R_numpy_322, tem=tem_21)

    edata.X = X_numpy_32
    _assert_shape_matches(edata, (3, 2, 2))
    assert edata.tem.shape == (2, 1)


#################################################################
### Test illegal initializations
#################################################################
def test_ehrdata_init_fail_r_2D(X_numpy_32):
    with pytest.raises(ValueError):
        EHRData(R=X_numpy_32)


def test_ehrdata_X_assign_fail_r_2D(X_numpy_32):
    edata = EHRData(X=X_numpy_32)
    with pytest.raises(ValueError):
        edata.R = X_numpy_32


def test_ehrdata_init_fail_X_and_R_mismatch(X_numpy_32, obs_31, var_21):
    r = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

    with pytest.raises(ValueError):
        EHRData(X=X_numpy_32, obs=obs_31, R=r, var=var_21)


def test_ehrdata_init_fail_R_and_t_mismatch(X_numpy_32, R_numpy_322, obs_31, var_21):
    tem = pd.DataFrame({"tem1": [1]})

    with pytest.raises(ValueError):
        EHRData(X=X_numpy_32, obs=obs_31, R=R_numpy_322, tem=tem, var=var_21)


def test_ehrdata_init_fail_R_and_layer_R_LAYER_KEY_exists(X_numpy_32, R_numpy_322):
    layers = {R_LAYER_KEY: R_numpy_322}

    with pytest.raises(ValueError):
        EHRData(X=X_numpy_32, R=R_numpy_322, layers=layers)


#################################################################
### Test t is protected alike obs, var
#################################################################
def test_ehrdata_set_aligneddataframes(X_numpy_32):
    edata_Xonly = EHRData(X_numpy_32)

    # show that setting df behavior for t alike obs, var
    with pytest.raises(ValueError):
        edata_Xonly.obs = pd.DataFrame([0, 1, 2, 3, 4])
    with pytest.raises(ValueError):
        edata_Xonly.var = pd.DataFrame([0, 1, 2, 3, 4])
    with pytest.raises(ValueError):
        edata_Xonly.tem = pd.DataFrame([0, 1, 2, 3, 4])


def test_ehrdata_del_r(X_numpy_32, R_numpy_322):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322)
    del edata.R

    assert edata.tem.shape == (0, 0)

    _assert_shape_matches(edata, (3, 2, 0), check_R_None=True)


#################################################################
### Test different types for R
#################################################################
# "R_sparse_322" currently disabled as sparse.COO support in AnnData is not yet implemented
@pytest.mark.parametrize("R_fixture_name", ["R_numpy_322", "R_dask_322"])
def test_ehrdata_R_data_types(R_fixture_name, request):
    R = request.getfixturevalue(R_fixture_name)
    edata = EHRData(R=R)
    _assert_shape_matches(edata, (3, 2, 2))


#################################################################
### Test sliceing
#################################################################
def test_ehrdata_subset_slice_2D_vanilla(X_numpy_32, R_numpy_322, obs_31, var_21):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322, obs=obs_31, var=var_21)
    edata_sliced = edata[:2, :1]

    assert edata_sliced.is_view
    _assert_shape_matches(edata_sliced, (2, 1, 2))


def test_ehrdata_subset_slice_2D_repeated(X_numpy_32, R_numpy_322, obs_31, var_21):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322, obs=obs_31, var=var_21)
    edata_sliced = edata[1:]
    edata_sliced = edata_sliced[1:]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (1, 2, 2))

    assert np.array_equal(edata_sliced.X, X_numpy_32[2].reshape(-1, 2))
    assert np.array_equal(edata_sliced.R, R_numpy_322[2].reshape(-1, 2, 2))


def test_ehrdata_subset_slice_3D_vanilla(X_numpy_32, R_numpy_322, obs_31, var_21, tem_21):
    edata = EHRData(X=X_numpy_32, R=R_numpy_322, obs=obs_31, var=var_21, tem=tem_21)
    edata_sliced = edata[:2, :1, :1]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (2, 1, 1))

    assert edata_sliced.tem.shape == (1, 1)


def test_ehrdata_subset_slice_3D_repeated(edata_333):
    edata = edata_333
    edata_sliced = edata[1:, 1:, 1:]
    edata_sliced = edata_sliced[1:, 1:, 1:]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (1, 1, 1))

    assert edata_sliced.tem.shape == (1, 1)

    # test that the true values are conserved
    assert np.array_equal(edata_sliced.X, edata.X[2, 2].reshape(-1, 1))
    assert np.array_equal(edata_sliced.R, edata.R[2, 2, 2].reshape(-1, 1, 1))
    assert np.array_equal(edata.tem.iloc[2].values.reshape(-1, 1), edata_sliced.tem.values)


def test_ehrdata_subset_obsvar_names_vanilla(edata_333, adata_33):
    edata = edata_333
    edata_a = edata[["obs1"]]
    _assert_fields_are_view(edata_a)
    _assert_shape_matches(edata_a, (1, 3, 3))

    edata_ab = edata[["obs1", "obs2"]]
    _assert_fields_are_view(edata_ab)
    _assert_shape_matches(edata_ab, (2, 3, 3))

    edata_a = edata[:, ["var1"]]
    _assert_fields_are_view(edata_a)
    _assert_shape_matches(edata_a, (3, 1, 3))

    edata_ab = edata[:, ["var1", "var2"]]
    _assert_fields_are_view(edata_ab)
    _assert_shape_matches(edata_ab, (3, 2, 3))

    edata_a = edata[["obs1", "obs2"], ["var1", "var2"]]
    _assert_fields_are_view(edata_a)
    _assert_shape_matches(edata_a, (2, 2, 3))
    assert edata_a.R.shape == (2, 2, 3)


def test_ehrdata_subset_obsvar_names_repeated(edata_333):
    edata = edata_333
    edata_ab = edata[["obs2", "obs3"]]
    edata_a = edata_ab[["obs3"]]

    _assert_fields_are_view(edata_a)
    _assert_shape_matches(edata_a, (1, 3, 3))

    assert np.array_equal(edata_a.R, edata.R[2, :, :].reshape(-1, 3, 3))

    edata = edata_333
    edata_ab = edata[:, ["var2", "var3"]]
    edata_a = edata_ab[:, ["var3"]]

    _assert_fields_are_view(edata_a)
    _assert_shape_matches(edata_a, (3, 1, 3))

    assert np.array_equal(edata_a.R, edata.R[:, 2, :].reshape(3, -1, 3))


def test_ehrdata_subset_boolindex_vanilla(edata_333):
    edata = edata_333
    edata_sliced = edata[[False, True, True], [True, False, False], [True, False, False]]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (2, 1, 1))

    assert edata_sliced.tem.shape == (1, 1)


def test_ehrdata_subset_boolindex_repeated(edata_333):
    edata = edata_333
    edata_sliced = edata[[False, True, True], [False, True, True], [False, True, True]]
    edata_sliced = edata_sliced[[False, True], [False, True], [False, True]]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (1, 1, 1))

    assert edata_sliced.tem.shape == (1, 1)

    # test that the true values are conserved
    assert np.array_equal(edata_sliced.X, edata.X[2, 2].reshape(-1, 1))
    assert np.array_equal(edata_sliced.R, edata.R[2, 2, 2].reshape(-1, 1, 1))
    assert np.array_equal(edata.tem.iloc[2].values.reshape(-1, 1), edata_sliced.tem.values)


def test_ehrdata_subset_numberindex_vanilla(edata_333):
    edata = edata_333
    edata_sliced = edata[[1, 2], [1], [1]]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (2, 1, 1))

    assert edata_sliced.tem.shape == (1, 1)


def test_ehrdata_subset_numberindex_repeated(edata_333):
    edata = edata_333
    edata_sliced = edata[[1, 2], [1, 2], [1, 2]]
    edata_sliced = edata_sliced[[1], [1], [1]]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (1, 1, 1))

    assert edata_sliced.tem.shape == (1, 1)

    # test that the true values are conserved
    assert np.array_equal(edata_sliced.X, edata.X[2, 2].reshape(-1, 1))
    assert np.array_equal(edata_sliced.R, edata.R[2, 2, 2].reshape(-1, 1, 1))
    assert np.array_equal(edata.tem.iloc[2].values.reshape(-1, 1), edata_sliced.tem.values)


def test_ehrdata_subset_mixedindices(edata_333):
    edata = edata_333

    edata_sliced = edata[["obs1", "obs2"], 1, [False, True, True]]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (2, 1, 2))

    edata_sliced = edata_sliced[np.array([False, True]), :, [1]]

    _assert_fields_are_view(edata_sliced)
    _assert_shape_matches(edata_sliced, (1, 1, 1))


def test_copy(edata_333):
    edata = EHRData()
    edata.copy()

    edata = edata_333
    edata_copy = edata.copy()

    assert isinstance(edata_copy.X, np.ndarray)
    assert isinstance(edata_copy.R, np.ndarray)
    assert isinstance(edata_copy.obs, pd.DataFrame)
    assert isinstance(edata_copy.var, pd.DataFrame)
    assert isinstance(edata_copy.tem, pd.DataFrame)

    assert edata.tem.shape == (3, 1)


def test_copy_of_slice(edata_333):
    edata = edata_333
    edata_sliced = edata[1:, 1:, 1:]
    edata_sliced_copy = edata_sliced.copy()

    _assert_shape_matches(edata_sliced_copy, (2, 2, 2))
    assert isinstance(edata_sliced_copy.X, np.ndarray)
    assert isinstance(edata_sliced_copy.R, np.ndarray)
    assert isinstance(edata_sliced_copy.obs, pd.DataFrame)
    assert isinstance(edata_sliced_copy.var, pd.DataFrame)
    assert isinstance(edata_sliced_copy.tem, pd.DataFrame)

    assert edata_sliced_copy.tem.shape == (2, 1)


def test_copy_of_obsvar_names(edata_333, adata_33):
    edata = edata_333

    edata_obs_subset = edata[["obs1", "obs2"]]
    edata_obs_subset = edata_obs_subset.copy()
    assert not edata_obs_subset.is_view
    _assert_shape_matches(edata_obs_subset, (2, 3, 3))

    edata_var_subset = edata[:, ["var1", "var2"]]
    edata_var_subset = edata_var_subset.copy()
    assert not edata_var_subset.is_view
    _assert_shape_matches(edata_var_subset, (3, 2, 3))

    edata_obsvar_subset = edata[["obs1"], ["var1", "var2"]]
    edata_obsvar_subset = edata_obsvar_subset.copy()
    assert not edata_obsvar_subset.is_view
    _assert_shape_matches(edata_obsvar_subset, (1, 2, 3))



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import pandas as pd

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from ehrdata import EHRData


def read_csv(
    filename: Path | str,
    *,
    sep: str = ",",
    index_column: str | None = None,
    columns_obs_only: Iterable[str] | None = None,
    format: Literal["flat", "wide", "long"] = "flat",
    wide_format_time_suffix: str | None = None,
    long_format_keys: dict[Literal["observation_column", "variable_column", "time_column", "value_column"], str]
    | None = None,
    **kwargs,
) -> EHRData:
    """Read a comma-separated values (csv) file into an :class:`~ehrdata.EHRData` object.

    It first reads the csv file using :func:`pandas.read_csv`, and then passes the resulting :class:`~pandas.DataFrame` to :func:`ehrdata.io.from_pandas`.
    See the documentation of :func:`ehrdata.io.from_pandas` for more details of table layouts.

    Args:
        filename: Path to the file or directory to read. Delegates to :func:`pandas.read_csv`.
        sep: Separator in the file. Delegates to :func:`pandas.read_csv`.
        index_column: If specified, this column of the csv file will be used for the `.obs` dataframe.
            Delegates to :func:`~ehrdata.io.from_pandas`.
        columns_obs_only: These columns will be added to the `.obs` dataframe only.
            Delegates to :func:`~ehrdata.io.from_pandas`.
        format: The format of the input dataframe. If the data is not longitudinal, choose `format="flat"`.
            If the data is longitudinal in the long format, choose `format="long"`.
            If the data is longitudinal in a wide format, choose `format="wide"`.
            Delegates to :func:`~ehrdata.io.from_pandas`.
        wide_format_time_suffix: Use only if `format="wide"`.
            Suffices in the variable columns that indicate the time of the observation.
            The collected suffices will be sorted lexicographically, and the variables ordered accordingly along the 3rd axis of the :class:`~ehrdata.EHRData` object.
            Delegates to :func:`~ehrdata.io.from_pandas`.
        long_format_keys: Use only if `format="long"`.
            The keys of the dataframe in the long format.
            The dictionary should have the following structure:
            {"observation_column": "<the column name of the observation ids>",
            "variable_column": "<the column name of the variable ids>",
            "time_column": "<the column name of the time>",
            "value_column": "<the column name of the values>"}.
            Delegates to :func:`~ehrdata.io.from_pandas`.
        **kwargs: Passed to :func:`pandas.read_csv`.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.io.read_csv("myfile.csv")
    """
    from ehrdata.io import from_pandas

    df = pd.read_csv(filename, sep=sep, index_col=index_column, **kwargs)
    edata = from_pandas(
        df,
        columns_obs_only=columns_obs_only,
        format=format,
        wide_format_time_suffix=wide_format_time_suffix,
        long_format_keys=long_format_keys,
    )

    return edata



from .datasets import (
    diabetes_130_fairlearn,
    diabetes_130_raw,
    ehrdata_blobs,
    gibleed_omop,
    mimic_2,
    mimic_2_preprocessed,
    mimic_iv_omop,
    physionet2012,
    synthea27nj_omop,
)

__all__ = [
    "diabetes_130_fairlearn",
    "diabetes_130_raw",
    "ehrdata_blobs",
    "gibleed_omop",
    "mimic_2",
    "mimic_2_preprocessed",
    "mimic_iv_omop",
    "physionet2012",
    "synthea27nj_omop",
]



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from sphinx.application import Sphinx


def skip_member_handler(
    app: Sphinx,
    what: Literal["module", "class", "exception", "function", "method", "attribute", "property"],
    name: str,
    obj: object,
    skip: bool,  # noqa: FBT001
    options,
) -> bool | None:
    """Skip inherited members."""
    if what not in {"method", "attribute", "property"}:
        return None
    if isinstance(obj, property):
        obj = obj.fget
    if name == "__getitem__":
        return False
    if name.startswith("_"):
        return True
    if not hasattr(obj, "__module__"):
        return None
    if not obj.__module__.startswith("ehrdata"):
        return True
    return None


def setup(app: Sphinx) -> None:
    """Setup lamindb for CI."""
    import lamindb as ln

    with suppress(RuntimeError):
        ln.setup.init(storage="/tmp/lamindb")

    app.connect("autodoc-skip-member", skip_member_handler)



from __future__ import annotations

import shutil
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from lamin_utils import logger

from ehrdata.core.constants import DEFAULT_DATA_PATH
from ehrdata.dt._dataloader import _download
from ehrdata.io import read_csv, read_h5ad
from ehrdata.io.omop import setup_connection
from ehrdata.io.omop._queries import _generate_timedeltas

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from duckdb import DuckDBPyConnection

    from ehrdata import EHRData


from scipy.sparse import csr_matrix
from sparse import COO


def ehrdata_blobs(
    *,
    n_variables: int = 11,
    n_centers: int = 5,
    cluster_std: float = 1.0,
    n_observations: int = 1000,
    base_timepoints: int = 100,
    random_state: int | np.random.Generator = 0,
    sparse: bool = False,
    sparsity: float = 0.9,
    variable_length: bool = False,
    time_shifts: bool = False,
    seasonality: bool = False,
    irregular_sampling: bool = False,
    missing_values: float = 0.0,
) -> EHRData:
    """Generates time series example dataset suited for alignment tasks.

    Args:
        n_variables: Dimension of feature space.
        n_centers: Number of cluster centers.
        cluster_std: Standard deviation of clusters.
        n_observations: Number of observations.
        base_timepoints: Base number of time points (actual may vary per observation).
        random_state: Determines random number generation for dataset creation.
        sparse: Whether to use sparse matrices.
        sparsity: Target sparsity level when sparse=True.
        variable_length: Whether observations have different time series lengths.
        time_shifts: Whether to add time shifts between similar observations.
        seasonality: Whether to add seasonal patterns to time series.
        irregular_sampling: Whether sampling intervals vary between observations.
        missing_values: Fraction of random missing values in time series.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.ehrdata_blobs(
        ...     variable_length=True, time_shifts=True, seasonality=True, irregular_sampling=True
        ... )

    results in a dataset like:

    .. image:: /_static/tutorial_images/ehrdata_blobs.png
       :alt: EHR data blobs visualization
    """
    rng = np.random.default_rng(random_state if isinstance(random_state, int) else None)

    # Generate cluster centers and assignments
    centers = rng.normal(0, 5, size=(n_centers, n_variables))
    y = rng.integers(0, n_centers, size=n_observations)

    # Generate base feature values (X)
    X = np.zeros((n_observations, n_variables))
    X = centers[y] + rng.normal(0, cluster_std, size=(n_observations, n_variables))

    # Determine time series lengths for each observation
    if variable_length:
        # Vary length from 50% to 150% of base_timepoints
        lengths = rng.integers(max(10, int(base_timepoints * 0.5)), int(base_timepoints * 1.5), size=n_observations)
    else:
        lengths = np.full(n_observations, base_timepoints)

    max_length = int(lengths.max())

    # Create time points for each observation (potentially irregular)
    all_timepoints = []
    for i in range(n_observations):
        length = lengths[i]

        if irregular_sampling:
            # Create non-uniform time spacing
            if time_shifts and i > 0:
                # Add random shift for similar clusters
                shift = rng.uniform(0, 0.3 * base_timepoints)
                start = shift if y[i] == y[i - 1] else 0
            else:
                start = 0

            # Irregular intervals with increasing spacing
            intervals = rng.exponential(scale=1.0, size=length)
            intervals = intervals / intervals.sum() * (max_length - start)
            timepoints = np.cumsum(intervals) + start
        else:
            # Regular intervals
            if time_shifts and i > 0:  # noqa: SIM108
                # Add cluster-based shifts
                shift = rng.uniform(0, 0.3 * base_timepoints) if y[i] == y[i - 1] else 0
            else:
                shift = 0

            timepoints = np.linspace(shift, max_length + shift, length)

        all_timepoints.append(timepoints)

    # Create time index - use all unique timepoints from all observations
    all_unique_times = np.unique(np.concatenate(all_timepoints))
    all_unique_times.sort()

    n_total_timepoints = len(all_unique_times)
    t_index = pd.Index([str(i) for i in range(n_total_timepoints)])

    # Create time DataFrame with actual time values
    t_df = pd.DataFrame(
        {
            "timepoint": range(n_total_timepoints),
            "time_value": all_unique_times,
        },
        index=t_index,
    )

    # Prepare 3D array to store all time series
    R = np.zeros((n_observations, n_variables, n_total_timepoints))
    R.fill(np.nan)

    # Generate time series for each observation
    for i in range(n_observations):
        # Map this observation's time points to the global time index
        obs_timepoints = all_timepoints[i]

        # Find indices of these timepoints in the global time array
        time_indices = np.searchsorted(all_unique_times, obs_timepoints)

        # Generate patterns for this observation
        for v in range(n_variables):
            base_value = X[i, v]

            # Time series with different patterns
            time_series = np.zeros(len(time_indices))

            # Add trend component (linear increase based on variable value)
            trend = np.linspace(0, base_value * 0.5, len(time_indices))
            time_series += trend

            # Add seasonality if enabled
            if seasonality:
                freq = rng.uniform(3, 15)

                # Phase shift based on cluster
                phase = y[i] * np.pi / n_centers

                # Amplitude based on variable value
                amplitude = np.abs(base_value) * 0.3

                # Add seasonal component
                seasonal = amplitude * np.sin(freq * np.pi * np.arange(len(time_indices)) / len(time_indices) + phase)
                time_series += seasonal

            # Add noise increasing with time
            for t_idx, t in enumerate(time_indices):
                noise_scale = cluster_std / 2 * (0.5 + t / n_total_timepoints)
                time_series[t_idx] += rng.normal(0, noise_scale)

            time_series += base_value

            R[i, v, time_indices] = time_series

    # Add random missing values if requested
    if missing_values > 0:
        # Create a mask for random missing values (ignoring already missing values)
        missing_mask = rng.random(R.shape) < missing_values
        not_nan_mask = ~np.isnan(R)
        R[missing_mask & not_nan_mask] = np.nan

    # Ensure that X contains a snapshot of R at a common time index
    # Use the middle timepoint if available for each observation
    for i in range(n_observations):
        valid_times = ~np.isnan(R[i, 0, :])
        if np.any(valid_times):
            # Find middle timepoint for this observation
            valid_indices = np.where(valid_times)[0]
            mid_idx = valid_indices[len(valid_indices) // 2]

            # Update X to contain this snapshot
            for v in range(n_variables):
                X[i, v] = R[i, v, mid_idx]

    if sparse:
        mask_x = rng.random(X.shape) > sparsity
        data_x = X.copy()
        data_x[~mask_x] = 0
        X = csr_matrix(data_x)

        # For R, handle both NaN and sparsity
        # First replace NaN with 0 where we're keeping values
        mask_r = rng.random(R.shape) > sparsity
        R_copy = R.copy()
        R_copy[np.isnan(R)] = 0
        R_copy[~mask_r] = 0

        # Get coordinates and values for non-zero entries
        coords = np.where(R_copy != 0)
        values = R_copy[coords]

        R = COO(coords, values, shape=R.shape)

    from ehrdata import EHRData

    return EHRData(
        X=X,
        obs=pd.DataFrame({"cluster": y.astype(str)}, index=pd.Index([str(i) for i in range(n_observations)])),
        var=pd.DataFrame(index=pd.Index([f"feature_{i}" for i in range(n_variables)])),
        R=R,
        tem=t_df,
    )


def _setup_eunomia_datasets(
    data_url: str,
    backend_handle: DuckDBPyConnection,
    data_path: Path,
    nested_omop_tables_folder: str | None = None,
    dataset_prefix: str = "",
) -> None:
    """Loads the Eunomia datasets in the OMOP Common Data model."""
    _download(
        data_url,
        output_path=data_path,
    )

    if nested_omop_tables_folder:
        if len(list((data_path / nested_omop_tables_folder).glob("*.csv"))) > 0:
            logger.info(f"Moving files from {data_path / nested_omop_tables_folder} to {data_path}")
        for file_path in (data_path / nested_omop_tables_folder).glob("*.csv"):
            shutil.move(file_path, data_path)

    setup_connection(
        data_path,
        backend_handle,
        prefix=dataset_prefix,
    )


def mimic_iv_omop(backend_handle: DuckDBPyConnection, data_path: Path | None = None) -> None:
    """Loads the MIMIC-IV demo data in the OMOP Common Data model.

    Loads the MIMIC-IV demo dataset from its `physionet repository <https://physionet.org/content/mimic-iv-demo-omop/0.9/#files-panel>`_ :cite:`kallfelz2021mimic`.

    Args:
        backend_handle: A handle to the backend which shall be used. Only duckdb connection supported at the moment.
        data_path: Path to the tables. If the path exists, the data is loaded from there. Else, the data is downloaded.

    Returns:
        Nothing. Adds the tables to the backend via the handle.

    Examples:
        >>> import ehrdata as ed
        >>> import duckdb
        >>> con = duckdb.connect()
        >>> ed.dt.mimic_iv_omop(backend_handle=con)
        >>> con.execute("SHOW TABLES;").fetchall()
    """
    data_url = "https://physionet.org/static/published-projects/mimic-iv-demo-omop/mimic-iv-demo-data-in-the-omop-common-data-model-0.9.zip"
    if data_path is None:
        data_path = DEFAULT_DATA_PATH / "ehrapy_data/mimic-iv-demo-data-in-the-omop-common-data-model-0.9"

    _setup_eunomia_datasets(
        data_url=data_url,
        backend_handle=backend_handle,
        data_path=data_path,
        nested_omop_tables_folder="mimic-iv-demo-data-in-the-omop-common-data-model-0.9/1_omop_data_csv",
        dataset_prefix="2b_",
    )


def gibleed_omop(backend_handle: DuckDBPyConnection, data_path: Path | None = None) -> None:
    """Loads the GiBleed dataset in the OMOP Common Data model.

    Loads the GIBleed dataset from the `EunomiaDatasets repository <https://github.com/OHDSI/EunomiaDatasets>`_.
    More details: https://github.com/OHDSI/EunomiaDatasets/tree/main/datasets/GiBleed.

    Args:
        backend_handle: A handle to the backend which shall be used. Only duckdb connection supported at the moment.
        data_path: Path to the tables. If the path exists, the data is loaded from there. Else, the data is downloaded.

    Returns:
        Nothing. Adds the tables to the backend via the handle.

    Examples:
        >>> import ehrdata as ed
        >>> import duckdb
        >>> con = duckdb.connect()
        >>> ed.dt.gibleed_omop(backend_handle=con)
        >>> con.execute("SHOW TABLES;").fetchall()
    """
    data_url = "https://github.com/OHDSI/EunomiaDatasets/raw/main/datasets/GiBleed/GiBleed_5.3.zip"

    if data_path is None:
        data_path = DEFAULT_DATA_PATH / "GiBleed_5.3"

    _setup_eunomia_datasets(
        data_url=data_url,
        backend_handle=backend_handle,
        data_path=data_path,
        nested_omop_tables_folder="GiBleed_5.3",
    )


def synthea27nj_omop(backend_handle: DuckDBPyConnection, data_path: Path | None = None) -> None:
    """Loads the Synthea27Nj dataset in the OMOP Common Data model.

    This function loads the Synthea27Nj dataset from the `EunomiaDatasets repository <https://github.com/OHDSI/EunomiaDatasets>`_.
    More details: https://github.com/OHDSI/EunomiaDatasets/tree/main/datasets/Synthea27Nj.

    Args:
        backend_handle: A handle to the backend which shall be used. Only duckdb connection supported at the moment.
        data_path: Path to the tables. If the path exists, the data is loaded from there. Else, the data is downloaded.

    Returns:
        Nothing. Adds the tables to the backend via the handle.

    Examples:
        >>> import ehrdata as ed
        >>> import duckdb
        >>> con = duckdb.connect()
        >>> ed.dt.synthea27nj_omop(backend_handle=con)
        >>> con.execute("SHOW TABLES;").fetchall()
    """
    data_url = "https://github.com/OHDSI/EunomiaDatasets/raw/main/datasets/Synthea27Nj/Synthea27Nj_5.4.zip"

    if data_path is None:
        data_path = DEFAULT_DATA_PATH / "Synthea27Nj_5.4"

    _setup_eunomia_datasets(
        data_url=data_url,
        backend_handle=backend_handle,
        data_path=data_path,
    )


def physionet2012(
    data_path: Path | None = None,
    *,
    interval_length_number: int = 1,
    interval_length_unit: str = "h",
    num_intervals: int = 48,
    aggregation_strategy: str = "last",
    drop_samples: Iterable[str] | None = [
        "147514",
        "142731",
        "145611",
        "140501",
        "155655",
        "143656",
        "156254",
        "150309",
        "140936",
        "141264",
        "150649",
        "142998",
    ],
) -> EHRData:
    """Loads the dataset of the `PhysioNet challenge 2012 (v1.0.0) <https://physionet.org/content/challenge-2012/1.0.0/>`_.

    If `interval_length_number` is 1, `interval_length_unit` is `"h"` (hour), and `num_intervals` is 48, this is the same as the `SAITS <https://arxiv.org/pdf/2202.08516>`_ preprocessing :cite:`du2023saits`.
    Truncated if a sample has more `num_intervals` steps; Padded if a sample has less than `num_intervals` steps.
    Further, by default the following 12 samples are dropped since they have no time series information at all: 147514, 142731, 145611, 140501, 155655, 143656, 156254, 150309,
    140936, 141264, 150649, 142998.
    Taken the defaults of `interval_length_number`, `interval_length_unit`, `num_intervals`, and `drop_samples`, the tensor stored in `.R` of `edata` is the same as when doing the `PyPOTS <https://github.com/WenjieDu/PyPOTS>`_ preprocessing :cite:`du2023pypots`.
    A simple deviation is that the tensor in `ehrdata` is of shape `n_obs x n_vars x n_intervals` (with defaults, 3000x37x48) while the tensor in PyPOTS is of shape `n_obs x n_intervals x n_vars` (3000x48x37).
    The tensor stored in `.R` is hence also fully compatible with the PyPOTS package, as the `.R` tensor of EHRData objects generally is.
    Note: In the original dataset, some missing values are encoded with a -1 for some entries of the variables `'DiasABP'`, `'NIDiasABP'`, and `'Weight'`. Here, these are replaced with `NaN` s.

    Args:
       data_path: Path to the raw data. If the path exists, the data is loaded from there.
           Else, the data is downloaded.
       interval_length_number: Numeric value of the length of one interval.
       interval_length_unit: Unit belonging to the interval length.
       num_intervals: Number of intervals.
       aggregation_strategy: Aggregation strategy for the time series data when multiple
           measurements for a person's parameter within a time interval is available.
           Available are `'first'` and `'last'`, as used in :meth:`~pandas.DataFrame.drop_duplicates`.
       drop_samples: Samples to drop from the dataset (indicate their RecordID).

    Returns:
        The processed physionet2012 dataset.
        The raw data is also downloaded, stored and available under the ``data_path``.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.physionet_2012()
    """
    from ehrdata import EHRData

    if data_path is None:
        data_path = DEFAULT_DATA_PATH / "physionet2012"

    # alternative in future?
    # config_parser = tsdb.utils.config.read_configs()
    # tsdb.utils.config.write_configs(
    #     config_parser=config_parser, key_value_set={"path": {"data_home": "ehrapy_data/physionet2012"}}
    # )

    # tsdb.load(dataset_name="physionet_2012")

    # # returns a dictionary
    # tsdb.data_processing.load_physionet2012(data_path)

    outcome_filenames = ["Outcomes-a.txt", "Outcomes-b.txt", "Outcomes-c.txt"]
    temp_data_set_names = ["set-a", "set-b", "set-c"]

    for filename in temp_data_set_names:
        _download(
            url=f"https://physionet.org/files/challenge-2012/1.0.0/{filename}.tar.gz?download",
            output_path=data_path,
            output_filename=f"{filename}.tar.gz",
            archive_format="tar.gz",
        )

    for filename in outcome_filenames:
        _download(
            url=f"https://physionet.org/files/challenge-2012/1.0.0/{filename}?download",
            output_path=data_path,
        )

    static_features = ["Age", "Gender", "ICUType", "Height"]

    person_long_across_set_collector = []
    for data_subset_dir in temp_data_set_names:
        person_long_within_set_collector = []

        # each txt file is the data of a person, in long format
        # the columns in the txt files are: Time, Parameter, Value
        for txt_file in (data_path / data_subset_dir).glob("*.txt"):
            person_long = pd.read_csv(txt_file)
            # drop the first row, which has the RecordID
            person_long = person_long.iloc[1:]

            # add RecordID (=person id in this dataset) to all data points of this person
            person_long["RecordID"] = int(txt_file.stem)
            person_long_within_set_collector.append(person_long)

        person_long_within_set_df = pd.concat(person_long_within_set_collector)

        person_long_within_set_df["set"] = data_subset_dir
        person_long_across_set_collector.append(person_long_within_set_df)

    person_long_across_set_df = pd.concat(person_long_across_set_collector)

    person_outcome_collector = []
    for outcome_filename in outcome_filenames:
        outcome_df = pd.read_csv(data_path / outcome_filename)
        person_outcome_collector.append(outcome_df)

    person_outcome_df = pd.concat(person_outcome_collector)

    # gather the static_features together with RecordID and set for each person into the obs table
    obs = (
        person_long_across_set_df[person_long_across_set_df["Parameter"].isin(static_features)]
        .pivot(index=["RecordID", "set"], columns=["Parameter"], values=["Value"])
        .reset_index(level="set", col_level=1)
    )
    obs.columns = obs.columns.droplevel(0)

    obs = obs.merge(person_outcome_df, how="left", left_on="RecordID", right_on="RecordID")
    obs.set_index("RecordID", inplace=True)

    # consider only time series features from now
    df_long = person_long_across_set_df[~person_long_across_set_df["Parameter"].isin(static_features)]

    interval_df = _generate_timedeltas(
        interval_length_number=interval_length_number,
        interval_length_unit=interval_length_unit,
        num_intervals=num_intervals,
    )

    df_long_time_seconds = np.array(pd.to_timedelta(df_long["Time"] + ":00").dt.total_seconds())
    interval_df_interval_end_offset_seconds = np.array(interval_df["interval_end_offset"].dt.total_seconds())
    df_long_interval_step = np.argmax(df_long_time_seconds[:, None] <= interval_df_interval_end_offset_seconds, axis=1)
    df_long.loc[:, ["interval_step"]] = df_long_interval_step

    # if one person for one feature (=Parameter) within one interval_step has multiple measurements, decide which one to keep
    df_long = df_long.drop_duplicates(subset=["RecordID", "Parameter", "interval_step"], keep=aggregation_strategy)

    xa = df_long.set_index(["RecordID", "Parameter", "interval_step"]).to_xarray()

    var = xa["Parameter"].to_dataframe()
    tem = xa["interval_step"].to_dataframe()
    r = xa["Value"].values

    # Three time series variables in the original dataset ['DiasABP', 'NIDiasABP', 'Weight'] have -1 instead of NaN for some missing values
    # No -1 value in the original dataset represents a valid measurement, so we can safely replace -1 with NaN
    r[r == -1] = np.nan

    obs.index = obs.index.astype(str)
    var.index = var.index.astype(str)

    edata = EHRData(R=r, obs=obs, var=var, tem=tem)

    return edata[~edata.obs.index.isin(drop_samples or [])]


def mimic_2(
    columns_obs_only: Iterable[str] | None = None,
) -> EHRData:
    """Loads the MIMIC-II dataset.

    This dataset was created for the purpose of a case study in the book: `Secondary Analysis of Electronic Health Records <https://link.springer.com/book/10.1007/978-3-319-43742-2>`_ :cite:`critical2016secondary`.
    In particular, the dataset was used to investigate the effectiveness of indwelling arterial catheters in hemodynamically stable patients with respiratory failure for mortality outcomes.
    The dataset is derived from MIMIC-II, the publicly-accessible critical care database.
    It contains summary clinical data and outcomes for 1,776 patients.

    More details on the data can be found on `physionet <https://physionet.org/content/mimic2-iaccd/1.0/>`_.

    Args:
        columns_obs_only: Columns to include only in obs and not X.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.mimic_2()
    """
    _download(
        "https://www.physionet.org/files/mimic2-iaccd/1.0/full_cohort_data.csv?download",
        output_path=DEFAULT_DATA_PATH,
        output_filename="ehrapy_mimic2.csv",
    )
    edata = read_csv(
        filename=f"{DEFAULT_DATA_PATH}/ehrapy_mimic2.csv",
        columns_obs_only=columns_obs_only,
    )

    return edata


def mimic_2_preprocessed() -> EHRData:
    """Loads the preprocessed MIMIC-II dataset.

    This dataset is a preprocessed version of :func:`~ehrdata.dt.mimic_2`.
    The dataset was preprocessed according to: https://github.com/theislab/ehrapy-datasets/tree/main/mimic_2.

    This dataset was created for the purpose of a case study in the book: `Secondary Analysis of Electronic Health Records <https://link.springer.com/book/10.1007/978-3-319-43742-2>`_ :cite:`critical2016secondary`.
    In particular, the dataset was used to investigate the effectiveness of indwelling arterial catheters in hemodynamically stable patients with respiratory failure for mortality outcomes.
    The dataset is derived from MIMIC-II, the publicly-accessible critical care database.
    It contains summary clinical data and outcomes for 1,776 patients.

    More details on the data can be found on `physionet <https://physionet.org/content/mimic2-iaccd/1.0/>`_.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.mimic_2_preprocessed()
    """
    _download(
        url="https://figshare.com/ndownloader/files/39727936",
        output_path=DEFAULT_DATA_PATH,
        output_filename="mimic_2_preprocessed.h5ad",
        raw_format="h5ad",
    )
    edata = read_h5ad(
        filename=f"{DEFAULT_DATA_PATH}/mimic_2_preprocessed.h5ad",
    )

    return edata


def diabetes_130_raw(
    columns_obs_only: Iterable[str] | None = None,
) -> EHRData:
    """Loads the raw diabetes-130 dataset.

    More details and the original dataset can be found `here <http://archive.ics.uci.edu/ml/datasets/Diabetes+130-US+hospitals+for+years+1999-2008>`_ :cite:`strack2014impact`.

    Args:
        columns_obs_only: Columns to include in `obs` only and not `X`.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.diabetes_130_raw()

    """
    _download(
        url="https://figshare.com/ndownloader/files/45110029",
        output_path=DEFAULT_DATA_PATH,
        output_filename="diabetes_130_raw.csv",
        raw_format="csv",
    )
    adata = read_csv(
        filename=f"{DEFAULT_DATA_PATH}/diabetes_130_raw.csv",
        columns_obs_only=columns_obs_only,
    )

    return adata


def diabetes_130_fairlearn(
    columns_obs_only: Iterable[str] | None = None,
) -> EHRData:
    """Loads the preprocessed diabetes-130 dataset by fairlearn.

    This loads the dataset from the `fairlearn.datasets.fetch_diabetes_hospital <https://fairlearn.org/v0.10/api_reference/generated/fairlearn.datasets.fetch_diabetes_hospital.html#fairlearn.datasets.fetch_diabetes_hospital>`_ function. :cite:`bird2020fairlearn`

    More details and the original dataset can be found `here <http://archive.ics.uci.edu/ml/datasets/Diabetes+130-US+hospitals+for+years+1999-2008>`_.

    Args:
        columns_obs_only: Columns to include in `obs` only and not `X`.

    Examples:
        >>> import ehrdata as ed
        >>> edata = ed.dt.diabetes_130_fairlearn()

    """
    _download(
        url="https://figshare.com/ndownloader/files/45110371",
        output_path=DEFAULT_DATA_PATH,
        output_filename="diabetes_130_fairlearn.csv",
        raw_format="csv",
    )
    edata = read_csv(
        filename=f"{DEFAULT_DATA_PATH}/diabetes_130_fairlearn.csv",
        columns_obs_only=columns_obs_only,
    )

    return edata
